"""Module for window openers."""

from .device import Device


class Window(Device):
    """Class for window openers."""

    def __init__(self, pyvlx, ident, name, subtype, typeid):
        """Initialize Window class."""
        self.pyvlx = pyvlx
        self.ident = ident
        self.name = name
        self.subtype = subtype
        self.typeid = typeid
        Device.__init__(self, pyvlx, ident, name)

    @classmethod
    def from_config(cls, pyvlx, item):
        """Read window opener from configuration."""
        name = item['name']
        ident = item['id']
        subtype = item['subtype']
        typeid = item['typeId']
        return cls(pyvlx, ident, name, subtype, typeid)

    def __str__(self) -> str:
        """Return object as readable string."""
        return '<Window name="{0}" ' \
            'id="{1}" ' \
            'subtype="{2}" ' \
            'typeId="{3}" />' \
            .format(
                self.name,
                self.ident,
                self.subtype,
                self.typeid)

    def __eq__(self, other):
        """Equal operator."""
        return self.__dict__ == other.__dict__



"""Module for house status monitor."""
from typing import TYPE_CHECKING

from .api_event import ApiEvent
from .frames import (
    FrameBase, FrameHouseStatusMonitorDisableConfirmation,
    FrameHouseStatusMonitorDisableRequest,
    FrameHouseStatusMonitorEnableConfirmation,
    FrameHouseStatusMonitorEnableRequest)

if TYPE_CHECKING:
    from pyvlx import PyVLX


class HouseStatusMonitorEnable(ApiEvent):
    """Class for enabling house status monotor."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize HouseStatusMonitorEnable class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameHouseStatusMonitorEnableConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self) -> FrameHouseStatusMonitorEnableRequest:
        """Construct initiating frame."""
        return FrameHouseStatusMonitorEnableRequest()


class HouseStatusMonitorDisable(ApiEvent):
    """Class for disabling house status monotor."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize HouseStatusMonitorEnable class."""
        super().__init__(pyvlx=pyvlx)
        self.success = False

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if not isinstance(frame, FrameHouseStatusMonitorDisableConfirmation):
            return False
        self.success = True
        return True

    def request_frame(self) -> FrameHouseStatusMonitorDisableRequest:
        """Construct initiating frame."""
        return FrameHouseStatusMonitorDisableRequest()



"""Unit tests for FrameGetSceneListNotification."""
import unittest

from pyvlx import PyVLXException
from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetSceneListNotification


class TestFrameGetSceneListNotification(unittest.TestCase):
    """Test class for FrameGetSceneListNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME1 = (
        b"\x00\xc8\x04\x0e\x03\x00All Window"
        + b"s Closed\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x01Sleeping "
        + b"Wide Open\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x02Bath Ope"
        + b"n\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x03\xe2"
    )

    EXAMPLE_FRAME2 = (
        b"\x00F\x04\x0e\x01\x00One Scene\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00w"
    )

    def test_bytes(self):
        """Test FrameGetSceneListNotification."""
        frame = FrameGetSceneListNotification()
        frame.scenes = [
            (0, "All Windows Closed"),
            (1, "Sleeping Wide Open"),
            (2, "Bath Open"),
        ]
        frame.remaining_scenes = 3
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME1)

    def test_bytes_one_scene(self):
        """Test FrameGetSceneListNotification with one scene."""
        frame = FrameGetSceneListNotification()
        frame.scenes = [(0, "One Scene")]
        frame.remaining_scenes = 0
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME2)

    def test_frame_from_raw(self):
        """Test parse FrameGetSceneListNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME1)
        self.assertTrue(isinstance(frame, FrameGetSceneListNotification))
        self.assertEqual(
            frame.scenes,
            [(0, "All Windows Closed"), (1, "Sleeping Wide Open"), (2, "Bath Open")],
        )
        self.assertEqual(frame.remaining_scenes, 3)

    def test_frame_from_raw_one_scene(self):
        """Test parse FrameGetSceneListNotification from raw with one scene."""
        frame = frame_from_raw(self.EXAMPLE_FRAME2)
        self.assertTrue(isinstance(frame, FrameGetSceneListNotification))
        self.assertEqual(frame.scenes, [(0, "One Scene")])
        self.assertEqual(frame.remaining_scenes, 0)

    def test_str(self):
        """Test string representation of FrameGetSceneListNotification."""
        frame = FrameGetSceneListNotification()
        frame.scenes = [
            (0, "All Windows Closed"),
            (1, "Sleeping Wide Open"),
            (2, "Bath Open"),
        ]
        frame.remaining_scenes = 3
        self.assertEqual(
            str(frame),
            '<FrameGetSceneListNotification scenes="[(0, \'All Windows Closed\'), '
            + '(1, \'Sleeping Wide Open\'), (2, \'Bath Open\')]" remaining_scenes="3">',
        )

    def test_wrong_payload(self):
        """Test wrong payload length, 2 scenes in len, only one provided."""
        frame = FrameGetSceneListNotification()
        with self.assertRaises(PyVLXException) as ctx:
            frame.from_payload(
                b"\x02\x00XXX\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            )
        self.assertEqual(
            ctx.exception.description, "scene_list_notification_wrong_length"
        )



"""Unit tests for PyVLX PasswordChangeNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FramePasswordChangeNotification
from pyvlx.exception import PyVLXException


class TestFramePasswordChange(unittest.TestCase):
    """Test class for FramePasswordChangeNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FramePasswordChangeNotification."""
        frame = FramePasswordChangeNotification(newpassword="fnord")
        self.assertEqual(
            bytes(frame),
            b"\x00#0\x04fnord\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00f",
        )

    def test_bytes_long_pw(self):
        """Test FramePasswordChangeNotification with long new password."""
        frame = FramePasswordChangeNotification(newpassword="x" * 32)
        self.assertEqual(
            bytes(frame), b"\x00#0\x04xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\x17"
        )

    def test_frame_from_raw(self):
        """Test parsing FramePasswordChangeNotification from raw bytes."""
        frame = frame_from_raw(
            b"\x00#0\x04fnord\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00f",
        )
        self.assertTrue(isinstance(frame, FramePasswordChangeNotification))
        self.assertEqual(frame.newpassword, "fnord")

    def test_errors(self):
        """Test FramePasswordChangeNotification with wrong password."""
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeNotification())
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeNotification(newpassword="x" * 33))

    def test_str(self):
        """Test string representation of FramePasswordChangeNotification."""
        frame = FramePasswordChangeNotification(newpassword="fnord")
        self.assertEqual(str(frame), '<FramePasswordChangeNotification newpassword="fn****"/>')

    def test_str_no_password(self):
        """Test string representation of FramePasswordChangeNotification with no password."""
        frame = FramePasswordChangeNotification()
        self.assertEqual(str(frame), '<FramePasswordChangeNotification newpassword="None"/>')



"""
Module for PyVLX object.

PyVLX is an asynchronous library for connecting to
a VELUX KLF 200 device for controlling window openers
and roller shutters.
"""
import logging

from .config import Config
from .interface import Interface
from .devices import Devices
from .scenes import Scenes


class PyVLX:
    """Class for PyVLX."""

    def __init__(self, path=None, host=None, password=None, timeout=10):
        """Initialize PyVLX class."""
        self.logger = logging.getLogger('pyvlx.log')
        self.config = Config(self, path, host, password)
        self.interface = Interface(self.config, timeout)
        self.devices = Devices(self)
        self.scenes = Scenes(self)

    async def connect(self):
        """Connect to KLF 200."""
        await self.interface.refresh_token()

    async def disconnect(self):
        """Disconnect from KLF 200."""
        await self.interface.disconnect()

    async def load_devices(self):
        """Load devices from KLF 200."""
        await self.devices.load()

    async def load_scenes(self):
        """Load scenes from KLF 200."""
        await self.scenes.load()



"""
Module for basic object for devices.

Device objectis an interface class and should
be derived by other objects like window openers
and roller shutters.
"""


class Device:
    """Class for device abstraction."""

    def __init__(self, pyvlx, ident, name):
        """Initialize Switch class."""
        self.pyvlx = pyvlx
        self.ident = ident
        self.name = name

    def get_name(self):
        """Return name of object."""
        return self.name



"""Unit tests for FrameSessionFinishedNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameSessionFinishedNotification


class TestFrameSessionFinishedNotification(unittest.TestCase):
    """Test class FrameSessionFinishedNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameSessionFinishedNotification."""
        frame = FrameSessionFinishedNotification(session_id=1000)
        self.assertEqual(bytes(frame), b"\x00\x05\x03\x04\x03\xe8\xe9")

    def test_frame_from_raw(self):
        """Test parse FrameSessionFinishedNotification from raw."""
        frame = frame_from_raw(b"\x00\x05\x03\x04\x03\xe8\xe9")
        self.assertTrue(isinstance(frame, FrameSessionFinishedNotification))
        self.assertEqual(frame.session_id, 1000)

    def test_str(self):
        """Test string representation of FrameSessionFinishedNotification."""
        frame = FrameSessionFinishedNotification(session_id=1000)
        self.assertEqual(
            str(frame), '<FrameSessionFinishedNotification session_id="1000"/>'
        )



"""Unit tests for FrameActivateSceneRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameActivateSceneRequest
from pyvlx.const import Originator, Priority, Velocity


class TestFrameActivateSceneRequest(unittest.TestCase):
    """Test class FrameActivateSceneRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\t\x04\x12\x03\xe8\x02\x03\x04\x01\xf0"

    def test_bytes(self):
        """Test FrameActivateSceneRequest with NO_TYPE."""
        frame = FrameActivateSceneRequest(
            scene_id=4,
            session_id=1000,
            originator=Originator.RAIN,
            velocity=Velocity.SILENT,
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameActivateSceneRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameActivateSceneRequest))
        self.assertEqual(frame.scene_id, 4)
        self.assertEqual(frame.session_id, 1000)
        self.assertEqual(frame.originator, Originator.RAIN)
        self.assertEqual(frame.priority, Priority.USER_LEVEL_2)
        self.assertEqual(frame.velocity, Velocity.SILENT)

    def test_str(self):
        """Test string representation of FrameActivateSceneRequest."""
        frame = FrameActivateSceneRequest(
            scene_id=4, session_id=1000, velocity=Velocity.FAST
        )
        self.assertEqual(
            str(frame),
            '<FrameActivateSceneRequest scene_id="4" session_id="1000" '
            'originator="Originator.USER" velocity="Velocity.FAST"/>',
        )



"""Unit tests for FramePasswordChangeConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FramePasswordChangeConfirmation, PasswordChangeConfirmationStatus)


class TestFramePasswordChangeConfirmation(unittest.TestCase):
    """Test class for FramePasswordChangeConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FramePasswordChangeConfirmation."""
        frame = FramePasswordChangeConfirmation()
        self.assertEqual(bytes(frame), b"\x00\x040\x03\x007")

    def test_bytes_error(self):
        """Test failed FramePasswordChangeConfirmation."""
        frame = FramePasswordChangeConfirmation(
            status=PasswordChangeConfirmationStatus.FAILED
        )
        self.assertEqual(bytes(frame), b"\x00\x040\x03\x016")

    def test_frame_from_raw(self):
        """Test parse FramePasswordChangeConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x040\x03\x016")
        self.assertTrue(isinstance(frame, FramePasswordChangeConfirmation))
        self.assertEqual(frame.status, PasswordChangeConfirmationStatus.FAILED)

    def test_str(self):
        """Test string representation of FramePasswordChangeConfirmation."""
        frame = FramePasswordChangeConfirmation()
        self.assertEqual(
            str(frame),
            '<FramePasswordChangeConfirmation status="PasswordChangeConfirmationStatus.SUCCESSFUL"/>',
        )



"""Unit tests for data sample obtained from MiSchroe."""
import unittest
from datetime import datetime

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FrameGetAllNodesInformationNotification,
    FrameGetNodeInformationNotification)
from pyvlx.const import (
    NodeTypeWithSubtype, NodeVariation, OperatingState, Velocity)
from pyvlx.parameter import Position
from pyvlx.slip import get_next_slip


class TestFrameGetNodeInformationMiSchroe(unittest.TestCase):
    """Test class data sample obtained from MiSchroe."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME1 = (
        "c0:00:7f:02:04:04:00:04:04:46:65:6e:73:74:65:72:"
        "20:42:c3:bc:72:6f:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:01:01:d5:07:00:01:"
        "1e:53:36:27:26:10:2f:00:81:05:c8:00:c8:00:00:00:"
        "f7:ff:f7:ff:f7:ff:00:00:4f:0d:f9:a7:02:d8:02:64:"
        "00:d8:03:ba:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:fb:c0"
    )

    EXAMPLE_FRAME2 = (
        "c0:00:7f:02:10:04:00:04:04:46:65:6e:73:74:65:72:"
        "20:42:c3:bc:72:6f:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:00:00:00:00:00:00:00:00:00:01:01:d5:07:00:01:"
        "1e:53:36:27:26:10:2f:00:81:05:c8:00:c8:00:00:00:"
        "f7:ff:f7:ff:f7:ff:00:00:4f:0d:f9:a8:02:d8:02:64:"
        "00:d8:03:ba:00:00:00:00:00:00:00:00:00:00:00:00:"
        "00:e0:c0"
    )

    def test_frame1_from_raw(self) -> None:
        """Test parse EXAMPLE_FRAME1 from raw."""
        slip = bytearray.fromhex(self.EXAMPLE_FRAME1.replace(":", ""))
        raw, _ = get_next_slip(slip)
        frame = frame_from_raw(bytes(raw))
        self.assertTrue(isinstance(frame, FrameGetAllNodesInformationNotification))
        self.assertEqual(frame.node_id, 4)
        self.assertEqual(frame.order, 4)
        self.assertEqual(frame.placement, 4)
        self.assertEqual(frame.name, "Fenster Büro")
        self.assertEqual(frame.velocity, Velocity.DEFAULT)
        self.assertEqual(
            frame.node_type, NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR
        )
        self.assertEqual(frame.product_group, 213)
        self.assertEqual(frame.product_type, 7)
        self.assertEqual(frame.node_variation, NodeVariation.NOT_SET)
        self.assertEqual(frame.power_mode, 1)
        self.assertEqual(frame.build_number, 30)
        self.assertEqual(frame.serial_number, "53:36:27:26:10:2f:00:81")
        self.assertEqual(frame.state, OperatingState.DONE)
        self.assertEqual(str(Position(frame.current_position)), "100 %")
        self.assertEqual(str(Position(frame.target)), "100 %")
        self.assertEqual(str(Position(frame.current_position_fp1)), "0 %")
        self.assertEqual(str(frame.current_position_fp2), "UNKNOWN")
        self.assertEqual(str(frame.current_position_fp3), "UNKNOWN")
        self.assertEqual(str(frame.current_position_fp4), "UNKNOWN")
        self.assertEqual(frame.remaining_time, 0)
        self.assertEqual(frame.timestamp, 1326315943)
        test_ts = datetime.fromtimestamp(1326315943).strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(frame.timestamp_formatted, test_ts)
        self.assertEqual(str(frame.alias_array), "d802=6400, d803=ba00")
        # Crosscheck, Serializing:
        self.assertEqual(bytes(frame), raw)

    def test_frame2_from_raw(self) -> None:
        """Test parse EXAMPLE_FRAME2 from raw."""
        slip = bytearray.fromhex(self.EXAMPLE_FRAME2.replace(":", ""))
        raw, _ = get_next_slip(slip)
        frame = frame_from_raw(bytes(raw))
        self.assertTrue(isinstance(frame, FrameGetNodeInformationNotification))
        self.assertEqual(frame.node_id, 4)
        self.assertEqual(frame.order, 4)
        self.assertEqual(frame.placement, 4)
        self.assertEqual(frame.name, "Fenster Büro")
        self.assertEqual(frame.velocity, Velocity.DEFAULT)
        self.assertEqual(
            frame.node_type, NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR
        )
        self.assertEqual(frame.product_group, 213)
        self.assertEqual(frame.product_type, 7)
        self.assertEqual(frame.node_variation, NodeVariation.NOT_SET)
        self.assertEqual(frame.power_mode, 1)
        self.assertEqual(frame.build_number, 30)
        self.assertEqual(frame.serial_number, "53:36:27:26:10:2f:00:81")
        self.assertEqual(frame.state, OperatingState.DONE)
        self.assertEqual(str(Position(frame.current_position)), "100 %")
        self.assertEqual(str(Position(frame.target)), "100 %")
        self.assertEqual(str(Position(frame.current_position_fp1)), "0 %")
        self.assertEqual(str(Position(frame.current_position_fp2)), "UNKNOWN")
        self.assertEqual(str(Position(frame.current_position_fp3)), "UNKNOWN")
        self.assertEqual(str(Position(frame.current_position_fp4)), "UNKNOWN")
        self.assertEqual(frame.remaining_time, 0)
        self.assertEqual(frame.timestamp, 1326315944)
        test_ts = datetime.fromtimestamp(1326315944).strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(frame.timestamp_formatted, test_ts)
        self.assertEqual(str(frame.alias_array), "d802=6400, d803=ba00")
        # Crosscheck, Serializing:
        self.assertEqual(bytes(frame), raw)



"""Module for handling the Reboot to API."""
from typing import TYPE_CHECKING

from pyvlx.log import PYVLXLOG

from .api_event import ApiEvent
from .frames import (
    FrameBase, FrameGatewayRebootConfirmation, FrameGatewayRebootRequest)

if TYPE_CHECKING:
    from pyvlx import PyVLX


class Reboot(ApiEvent):
    """Class for handling Reboot to API."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize Reboot class."""
        super().__init__(pyvlx=pyvlx)
        self.pyvlx = pyvlx
        self.success = False

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if isinstance(frame, FrameGatewayRebootConfirmation):
            PYVLXLOG.warning("KLF200 is rebooting")
            self.success = True
            return True
        return False

    def request_frame(self) -> FrameGatewayRebootRequest:
        """Construct initiating frame."""
        return FrameGatewayRebootRequest()



"""Unit test for lights."""
import unittest
from unittest.mock import MagicMock

from pyvlx import Light, PyVLX
from pyvlx.connection import Connection


# pylint: disable=too-many-public-methods,invalid-name
class TestDimmableDevice(unittest.TestCase):
    """Test class for lights."""

    def setUp(self) -> None:
        """Set up TestDimmableDevice."""
        self.pyvlx = MagicMock(spec=PyVLX)
        connection = MagicMock(spec=Connection)
        self.pyvlx.attach_mock(mock=connection, attribute="connection")

    def test_light_str(self) -> None:
        """Test string representation of Light object."""
        light = Light(
            pyvlx=self.pyvlx,
            node_id=23,
            name="Test Light",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(light),
            '<Light name="Test Light" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23"/>',
        )

    def test_eq(self) -> None:
        """Testing eq method with positive results."""
        node1 = Light(
            pyvlx=self.pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        node2 = Light(
            pyvlx=self.pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        self.assertEqual(node1, node2)

    def test_nq(self) -> None:
        """Testing eq method with negative results."""
        node1 = Light(
            pyvlx=self.pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        node2 = Light(
            pyvlx=self.pyvlx, node_id=24, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        self.assertNotEqual(node1, node2)



"""
Module for basic object for nodes.

Node object is an interface class and should
be derived by other objects like window openers
and roller shutters.
"""
from typing import TYPE_CHECKING, Any, Awaitable, Callable, List, Optional

from .api import SetNodeName
from .exception import PyVLXException

if TYPE_CHECKING:
    from pyvlx import PyVLX

CallbackType = Callable[["Node"], Awaitable[None]]


class Node:
    """Class for node abstraction."""

    def __init__(self, pyvlx: "PyVLX", node_id: int, name: str, serial_number: Optional[str]):
        """Initialize Node object."""
        self.pyvlx = pyvlx
        self.node_id = node_id
        self.name = name
        self.serial_number = serial_number
        self.device_updated_cbs: List[CallbackType] = []
        self.pyvlx.connection.register_connection_opened_cb(self.after_update)
        self.pyvlx.connection.register_connection_closed_cb(self.after_update)

    def register_device_updated_cb(self, device_updated_cb: CallbackType) -> None:
        """Register device updated callback."""
        self.device_updated_cbs.append(device_updated_cb)

    def unregister_device_updated_cb(self, device_updated_cb: CallbackType) -> None:
        """Unregister device updated callback."""
        self.device_updated_cbs.remove(device_updated_cb)

    async def after_update(self) -> None:
        """Execute callbacks after internal state has been changed."""
        for device_updated_cb in self.device_updated_cbs:
            # pylint: disable=not-callable
            await self.pyvlx.loop.create_task(device_updated_cb(self))  # type: ignore

    async def rename(self, name: str) -> None:
        """Change name of node."""
        set_node_name = SetNodeName(pyvlx=self.pyvlx, node_id=self.node_id, name=name)
        await set_node_name.do_api_call()
        if not set_node_name.success:
            raise PyVLXException("Unable to rename node")
        self.name = name

    @property
    def is_available(self) -> bool:
        """Return True if node is available."""
        return self.pyvlx.get_connected()

    def __str__(self) -> str:
        """Return object as readable string."""
        return (
            '<{} name="{}" '
            'node_id="{}" '
            'serial_number="{}"/>'.format(
                type(self).__name__, self.name, self.node_id, self.serial_number
            )
        )

    def __eq__(self, other: Any) -> bool:
        """Equal operator."""
        return (
            type(self).__name__ == type(other).__name__
            and self.__dict__ == other.__dict__
        )



"""Module for Position class."""
from typing import Optional

from .exception import PyVLXException


class Parameter:
    """General object for storing parameters."""

    UNKNOWN_VALUE = 0xF7FF  # F7 FF
    CURRENT = 0xD200  # D2 00
    MAX = 0xC800  # C8 00
    MIN = 0x0000  # 00 00
    ON = 0x0000  # 00 00
    OFF = 0xC800  # C8 00
    TARGET = 0xD100  # D1 00
    IGNORE = 0xD400  # D4 00
    DUAL_SHUTTER_CURTAINS = 0xD808  # D8 08

    def __init__(self, raw: Optional[bytes] = None):
        """Initialize Parameter class."""
        self.raw = self.from_int(Position.UNKNOWN_VALUE)
        if raw is not None:
            self.raw = self.from_raw(raw)

    def __bytes__(self) -> bytes:
        """Convert object in byte representation."""
        return self.raw

    def from_parameter(self, parameter: "Parameter") -> None:
        """Set internal raw state from parameter."""
        if not isinstance(parameter, Parameter):
            raise PyVLXException("parameter::from_parameter_wrong_object")
        self.raw = parameter.raw

    @staticmethod
    def from_int(value: int) -> bytes:
        """Create raw out of position value."""
        if not isinstance(value, int):
            raise PyVLXException("value_has_to_be_int")
        if not Parameter.is_valid_int(value):
            raise PyVLXException("value_out_of_range")
        return bytes([value >> 8 & 255, value & 255])

    @staticmethod
    def to_int(raw: bytes) -> int:
        """Create int position value out of raw."""
        return raw[0] * 256 + raw[1]

    @staticmethod
    def is_valid_int(value: int) -> bool:
        """Test if value can be rendered out of int."""
        if 0 <= value <= Parameter.MAX:  # This includes ON and OFF
            return True
        valid_values = {
            Parameter.UNKNOWN_VALUE,
            Parameter.IGNORE,
            Parameter.CURRENT,
            Parameter.TARGET,
            Parameter.DUAL_SHUTTER_CURTAINS,
        }
        if value in valid_values:
            return True
        return False

    @staticmethod
    def from_raw(raw: bytes) -> bytes:
        """Test if raw packets are valid for initialization of Position."""
        if not isinstance(raw, bytes):
            raise PyVLXException("Position::raw_must_be_bytes")
        if len(raw) != 2:
            raise PyVLXException("Position::raw_must_be_two_bytes")
        if (
            raw != Position.from_int(Position.CURRENT)
            and raw != Position.from_int(Position.IGNORE)
            and raw != Position.from_int(Position.TARGET)
            and raw != Position.from_int(Position.UNKNOWN_VALUE)
            and Position.to_int(raw) > Position.MAX
        ):
            return Position.from_int(Position.UNKNOWN_VALUE)
        return raw

    @staticmethod
    def from_percent(percent: int) -> bytes:
        """Create raw value out of percent position."""
        if not isinstance(percent, int):
            raise PyVLXException("Position::percent_has_to_be_int")
        if percent < 0:
            raise PyVLXException("Position::percent_has_to_be_positive")
        if percent > 100:
            raise PyVLXException("Position::percent_out_of_range")
        return bytes([percent * 2, 0])

    @staticmethod
    def to_percent(raw: bytes) -> int:
        """Create percent position value out of raw."""
        # The first byte has the vlue from 0 to 200. Ignoring the second one.
        # Adding 0.5 allows a slight tolerance for devices (e.g. Velux SML) that
        # do not return exactly 51200 as final position when closed.
        return int(raw[0] / 2 + 0.5)

    def __eq__(self, other: object) -> bool:
        """Equal operator."""
        if not isinstance(other, Parameter):
            return NotImplemented
        return self.raw == other.raw

    def __str__(self) -> str:
        """Return string representation of object."""
        if self.raw == self.from_int(Position.UNKNOWN_VALUE):
            return "UNKNOWN"
        if self.raw == self.from_int(Position.CURRENT):
            return "CURRENT"
        if self.raw == self.from_int(Position.TARGET):
            return "TARGET"
        if self.raw == self.from_int(Position.IGNORE):
            return "IGNORE"
        if self.raw == self.from_int(Position.DUAL_SHUTTER_CURTAINS):
            return "DUAL"
        return "{} %".format(int(self.to_percent(self.raw)))


class SwitchParameter(Parameter):
    """Class for storing On or Off values."""

    def __init__(
        self, parameter: Optional[Parameter] = None, state: Optional[int] = None
    ):
        """Initialize Parameter class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)
        elif state is not None:
            self.state = state

    @property
    def state(self) -> int:
        """Position property."""
        return self.to_int(self.raw)

    @state.setter
    def state(self, state: int) -> None:
        """Setter of internal raw via state."""
        self.raw = self.from_int(state)

    def set_on(self) -> None:
        """Set parameter to 'on' state."""
        self.raw = self.from_int(Parameter.ON)

    def set_off(self) -> None:
        """Set parameter to 'off' state."""
        self.raw = self.from_int(Parameter.OFF)

    def is_on(self) -> bool:
        """Return True if parameter is in 'on' state."""
        return self.raw == self.from_int(Parameter.ON)

    def is_off(self) -> bool:
        """Return True if parameter is in 'off' state."""
        return self.raw == self.from_int(Parameter.OFF)

    def __str__(self) -> str:
        """Return string representation of object."""
        if self.raw == self.from_int(Parameter.ON):
            return "ON"
        if self.raw == self.from_int(Parameter.OFF):
            return "OFF"
        return "UNKNOWN"


class SwitchParameterOn(SwitchParameter):
    """Switch Parameter in switched 'on' state."""

    def __init__(self) -> None:
        """Initialize SwitchParameterOn class."""
        super().__init__(state=Parameter.ON)


class SwitchParameterOff(SwitchParameter):
    """Switch Parameter in switched 'off' state."""

    def __init__(self) -> None:
        """Initialize SwitchParameterOff class."""
        super().__init__(state=Parameter.OFF)


class Position(Parameter):
    """Class for storing a position."""

    def __init__(
        self,
        parameter: Optional[Parameter] = None,
        position: Optional[int] = None,
        position_percent: Optional[int] = None,
    ):
        """Initialize Position class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)
        elif position is not None:
            self.position = position
        elif position_percent is not None:
            self.position_percent = position_percent

    @property
    def known(self) -> bool:
        """Known property, true if position is not in an unknown position."""
        return self.raw != self.from_int(Position.UNKNOWN_VALUE)

    @property
    def open(self) -> bool:
        """Return true if position is set to fully open."""
        return self.raw == self.from_int(Position.MIN)

    @property
    def closed(self) -> bool:
        """Return true if position is set to fully closed."""
        # Consider closed even if raw is not exactly 51200 (tolerance for devices like Velux SML)
        return self.to_percent(self.raw) == self.to_percent(self.from_int(Position.MAX))

    @property
    def position(self) -> int:
        """Position property."""
        return self.to_int(self.raw)

    @position.setter
    def position(self, position: int) -> None:
        """Setter of internal raw via position."""
        self.raw = self.from_int(position)

    @property
    def position_percent(self) -> int:
        """Position percent property."""
        # unclear why it returns a <property object> here
        return int(self.to_percent(self.raw))

    @position_percent.setter
    def position_percent(self, position_percent: int) -> None:
        """Setter of internal raw via percent position."""
        self.raw = self.from_percent(percent=position_percent)


class UnknownPosition(Position):
    """Unknown position."""

    def __init__(self) -> None:
        """Initialize UnknownPosition class."""
        super().__init__(position=Position.UNKNOWN_VALUE)


class CurrentPosition(Position):
    """Current position, used to stop devices."""

    def __init__(self) -> None:
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.CURRENT)


class TargetPosition(Position):
    """Class for using a target position."""

    def __init__(self) -> None:
        """Initialize TargetPosition class."""
        super().__init__(position=Position.TARGET)


class IgnorePosition(Position):
    """The Ignore is used where a parameter in the frame is to be ignored."""

    def __init__(self) -> None:
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.IGNORE)


class Intensity(Parameter):
    """Class for storing an intensity, used in DimmableDevice.

    - 0% means off
    - 100% means fully on
    """

    def __init__(
        self,
        parameter: Optional[Parameter] = None,
        intensity: Optional[int] = None,
        intensity_percent: Optional[int] = None,
    ):
        """Initialize Intensity class."""
        super().__init__()
        if parameter is not None:
            self.from_parameter(parameter)
        elif intensity is not None:
            self.intensity = intensity
        elif intensity_percent is not None:
            self.intensity_percent = intensity_percent

    @property
    def known(self) -> bool:
        """Known property, true if intensity is not in an unknown intensity."""
        return self.raw != self.from_int(Intensity.UNKNOWN_VALUE)

    @property
    def on(self) -> bool:  # pylint: disable=invalid-name
        """Intensity at maximum power (fully on)."""
        return self.raw == self.from_int(Intensity.ON)

    @property
    def off(self) -> bool:
        """Intensity off state (no power)."""
        return self.raw == self.from_int(Intensity.OFF)

    @property
    def intensity(self) -> int:
        """Intensity property."""
        return self.to_int(self.raw)

    @intensity.setter
    def intensity(self, intensity: int) -> None:
        """Setter of internal raw via intensity."""
        self.raw = self.from_int(intensity)

    @staticmethod
    def from_percent(percent: int) -> bytes:
        """Create raw value out of percent intensity.

        For Intensity, 0% = off , 100% = fully on.
        This inverts the raw value: 0% -> 200, 100% -> 0.
        """
        if not isinstance(percent, int):
            raise PyVLXException("Intensity::percent_has_to_be_int")
        if percent < 0:
            raise PyVLXException("Intensity::percent_has_to_be_positive")
        if percent > 100:
            raise PyVLXException("Intensity::percent_out_of_range")
        # Invert: 0% = off (200), 100% = on (0)
        return bytes([(100 - percent) * 2, 0])

    @staticmethod
    def to_percent(raw: bytes) -> int:
        """Create percent intensity value out of raw.

        For Intensity, raw value 0 = 100% (fully on), raw value 200 = 0% (off).
        """
        # Invert: raw 0 = 100%, raw 200 = 0%
        return int(100 - raw[0] / 2 + 0.5)

    @property
    def intensity_percent(self) -> int:
        """Intensity percent property."""
        return int(self.to_percent(self.raw))

    @intensity_percent.setter
    def intensity_percent(self, intensity_percent: int) -> None:
        """Setter of internal raw via percent intensity."""
        self.raw = self.from_percent(percent=intensity_percent)

    def __str__(self) -> str:
        """Return string representation of object."""
        if self.raw == self.from_int(Intensity.UNKNOWN_VALUE):
            return "UNKNOWN"
        if self.raw == self.from_int(Intensity.CURRENT):
            return "CURRENT"
        if self.raw == self.from_int(Intensity.TARGET):
            return "TARGET"
        if self.raw == self.from_int(Intensity.IGNORE):
            return "IGNORE"
        return "{} %".format(self.intensity_percent)


class UnknownIntensity(Intensity):
    """Unknown intensity."""

    def __init__(self) -> None:
        """Initialize UnknownIntensity class."""
        super().__init__(intensity=Intensity.UNKNOWN_VALUE)


class CurrentIntensity(Intensity):
    """Current intensity, used to stop devices."""

    def __init__(self) -> None:
        """Initialize CurrentIntensity class."""
        super().__init__(intensity=Intensity.CURRENT)


class DualRollerShutterPosition(Position):
    """Position to be provided when addressing the upper or lower curtain of a dual roller shutter by using FP1 or FP2."""

    def __init__(self) -> None:
        """Initialize CurrentPosition class."""
        super().__init__(position=Position.DUAL_SHUTTER_CURTAINS)



"""Unit tests frame_creation module."""
import unittest

from pyvlx.api.frames import calc_crc, extract_from_frame
from pyvlx.exception import PyVLXException


class TestFrameHelper(unittest.TestCase):
    """Test class for frame_creation module."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_crc(self):
        """Test crc calculation."""
        self.assertEqual(calc_crc(b""), 0)
        self.assertEqual(calc_crc(b"\x01"), 1)
        self.assertEqual(calc_crc(b"\x01\x02"), 3)
        self.assertEqual(calc_crc(b"\x01\x02\x03"), 0)
        self.assertEqual(calc_crc(b"\x01\x02\x03\xff"), 255)

    def test_extract_from_frame(self):
        """Test extract_from_frame method."""
        with self.assertRaises(PyVLXException):
            extract_from_frame(bytes(4))
        with self.assertRaises(PyVLXException):
            extract_from_frame(
                bytes(b"\x01\x04\x00\x00\x02\x06")
            )  # invalid length (first 2 bytes)
        with self.assertRaises(PyVLXException):
            extract_from_frame(bytes(b"\x00\x04\x00\x00\x02\x07"))  # invalid crc
        with self.assertRaises(PyVLXException):
            extract_from_frame(bytes(b"\x00\x04\xFF\xFF\x02\x06"))  # invalid crc



"""Test for Position class."""
import unittest

from pyvlx import Parameter, Position
from pyvlx.exception import PyVLXException


class TestParameterPosition(unittest.TestCase):
    """Test class for Position class."""

    def test_from_parameter(self) -> None:
        """Test from parameter from another parameter."""
        param1 = Parameter(raw=b'\xC8\x00')
        param2 = Parameter(raw=b'\x00\x00')
        wrong_object = b'\x00\x00'
        with self.assertRaises(PyVLXException):
            Parameter.from_parameter(self=param1, parameter=wrong_object)  # type: ignore
        Parameter.from_parameter(self=param1, parameter=param2)
        self.assertEqual(param1.raw, param2.raw)

    def test_from_int(self) -> None:
        """Test from int output from parameter int input."""
        wrong_object = b'\x00\x00'
        with self.assertRaises(PyVLXException):
            Parameter.from_int(wrong_object)  # type: ignore
        not_valid_int = 100.3
        with self.assertRaises(PyVLXException):
            Parameter.from_int(not_valid_int)  # type: ignore
        int_out_of_range = Parameter.MAX + 1
        with self.assertRaises(PyVLXException):
            Parameter.from_int(int_out_of_range)  # type: ignore
        self.assertEqual(Parameter.from_int(51200), b'\xc8\x00')

    def test_to_int(self) -> None:
        """Test from int output from parameter int input."""
        raw = b'\xc8\x00'
        self.assertEqual(Parameter.to_int(raw=raw), 51200)

    def test_is_valid_int(self) -> None:
        """Test if int is valid, between 0 and 51200."""
        valid_int1 = 0
        valid_int2 = 51200
        out_of_range_int = 51201
        self.assertTrue(Parameter.is_valid_int(valid_int1))
        self.assertTrue(Parameter.is_valid_int(valid_int2))
        self.assertFalse(Parameter.is_valid_int(out_of_range_int))

    def test_from_raw(self) -> None:
        """Test raw output from byte input."""
        with self.assertRaises(PyVLXException):
            Parameter.from_raw(raw=0xC800)  # type: ignore
        with self.assertRaises(PyVLXException):
            Parameter.from_raw(raw=b'\x00')
        self.assertEqual(Parameter.from_raw(raw=b'\xd2\x01'), b'\xf7\xff')
        self.assertEqual(Parameter.from_raw(raw=b'\xd2\x01'), b'\xf7\xff')
        self.assertEqual(Parameter.from_raw(raw=b'\xd1\x01'), b'\xf7\xff')
        self.assertEqual(Parameter.from_raw(raw=b'\xd4\x01'), b'\xf7\xff')
        self.assertEqual(Parameter.from_raw(raw=b'\xd8\x01'), b'\xf7\xff')
        self.assertNotEqual(Parameter.from_raw(raw=b'\xd2\x00'), b'\xf7\xff')
        self.assertNotEqual(Parameter.from_raw(raw=b'\xd2\x00'), b'\xf7\xff')
        self.assertNotEqual(Parameter.from_raw(raw=b'\xd1\x00'), b'\xf7\xff')
        self.assertNotEqual(Parameter.from_raw(raw=b'\xd4\x00'), b'\xf7\xff')
        self.assertEqual(Parameter.from_raw(raw=b'\x00\x00'), b'\x00\x00')
        self.assertEqual(Parameter.from_raw(raw=b'\xc8\x00'), b'\xc8\x00')
        self.assertEqual(Parameter.from_raw(raw=b'\xc8\x01'), b'\xf7\xff')

    def test_from_to_percent(self) -> None:
        """Test position percent output from position percent input."""
        self.assertEqual(Position.to_percent(Position.from_percent(0)), 0)
        self.assertEqual(Position.to_percent(Position.from_percent(1)), 1)
        self.assertEqual(Position.to_percent(Position.from_percent(25)), 25)
        self.assertEqual(Position.to_percent(Position.from_percent(50)), 50)
        self.assertEqual(Position.to_percent(Position.from_percent(75)), 75)
        self.assertEqual(Position.to_percent(Position.from_percent(99)), 99)
        self.assertEqual(Position.to_percent(Position.from_percent(100)), 100)

    def test_to_percent(self) -> None:
        """Test position percent output from position int input."""
        self.assertEqual(Position.to_percent(Parameter.from_int(0)), 0)
        self.assertEqual(Position.to_percent(Parameter.from_int(1)), 0)
        self.assertEqual(Position.to_percent(Parameter.from_int(3)), 0)
        self.assertEqual(Position.to_percent(Parameter.from_int(256)), 1)
        self.assertEqual(Position.to_percent(Parameter.from_int(512)), 1)
        self.assertEqual(Position.to_percent(Parameter.from_int(5120)), 10)
        self.assertEqual(Position.to_percent(Parameter.from_int(12800)), 25)
        self.assertEqual(Position.to_percent(Parameter.from_int(25600)), 50)
        self.assertEqual(Position.to_percent(Parameter.from_int(38400)), 75)
        self.assertEqual(Position.to_percent(Parameter.from_int(50688)), 99)
        self.assertEqual(Position.to_percent(Parameter.from_int(51050)), 100)
        self.assertEqual(Position.to_percent(Parameter.from_int(51200)), 100)

    def test_equal(self) -> None:
        """Test from parameter from another parameter."""
        param1 = Parameter(raw=b'\xC8\x00')
        param2 = Parameter(raw=b'\x00\x00')
        param3 = Parameter(raw=b'\xC8\x00')
        wrong_object = b'\x00\x00'
        self.assertEqual(param1.__eq__(wrong_object), NotImplemented)  # pylint: disable=C2801
        self.assertFalse(param1 == param2)
        self.assertTrue(param1 == param3)



"""Unit tests _helper module."""
import unittest
from unittest.mock import MagicMock

from pyvlx import (
    Blade, Blind, GarageDoor, Gate, Light, OnOffLight, PyVLX, RollerShutter,
    Window)
from pyvlx.api.frames import FrameGetNodeInformationNotification
from pyvlx.connection import Connection
from pyvlx.const import NodeTypeWithSubtype
from pyvlx.dimmable_device import ExteriorHeating
from pyvlx.node_helper import convert_frame_to_node


class TestNodeHelper(unittest.TestCase):
    """Test class for helper functions of node_helper."""

    # pylint: disable=too-many-public-methods,invalid-name

    def setUp(self) -> None:
        """Set up TestNodeHelper."""
        self.pyvlx = MagicMock(spec=PyVLX)
        connection = MagicMock(spec=Connection)
        self.pyvlx.attach_mock(mock=connection, attribute="connection")

    def test_window(self) -> None:
        """Test convert_frame_to_node with window."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.WINDOW_OPENER
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Window(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_window_with_rain_sensor(self) -> None:
        """Test convert_frame_to_node with window with rain sensor."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Window(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                rain_sensor=True,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_blind(self) -> None:
        """Test convert_frame_to_node with blind."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.EXTERIOR_VENETIAN_BLIND
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Blind(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_roller_shutter(self) -> None:
        """Test convert_frame_to_node roller shutter."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.ROLLER_SHUTTER
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            RollerShutter(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_garage_door(self) -> None:
        """Test convert_frame_to_node garage door."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.GARAGE_DOOR_OPENER
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            GarageDoor(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_linar_angular_position_of_garage_door(self) -> None:
        """Test convert_frame_to_node linar angular position of garage door."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.LINEAR_ANGULAR_POSITION_OF_GARAGE_DOOR
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            GarageDoor(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_gate(self) -> None:
        """Test convert_frame_to_node gate."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.GATE_OPENER
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Gate(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_gate_with_angular_position(self) -> None:
        """Test convert_frame_to_node gate."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.GATE_OPENER_ANGULAR_POSITION
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Gate(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_blade(self) -> None:
        """Test convert_frame_to_node blade."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.BLADE_OPENER
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Blade(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_no_type(self) -> None:
        """Test convert_frame_to_node with no type (convert_frame_to_node should return None)."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.NO_TYPE
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        self.assertEqual(convert_frame_to_node(self.pyvlx, frame), None)

    def test_light(self) -> None:
        """Test convert_frame_to_node with light."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.LIGHT
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            Light(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_light_on_off(self) -> None:
        """Test convert_frame_to_node with light_on_off."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.LIGHT_ON_OFF
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            OnOffLight(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )

    def test_exterior_heating(self) -> None:
        """Test convert_frame_to_node with exterior heating."""
        frame = FrameGetNodeInformationNotification()
        frame.node_id = 23
        frame.name = "Fnord23"
        frame.node_type = NodeTypeWithSubtype.EXTERIOR_HEATING
        frame.serial_number = "aa:bb:aa:bb:aa:bb:aa:23"
        node = convert_frame_to_node(self.pyvlx, frame)
        self.assertEqual(
            node,
            ExteriorHeating(
                pyvlx=self.pyvlx,
                name="Fnord23",
                node_id=23,
                serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            ),
        )



"""Unit tests for FrameNodeStatePositionChangedNotification."""
import unittest
from datetime import datetime

from pyvlx import Position
from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameNodeStatePositionChangedNotification
from pyvlx.const import OperatingState


class TestFrameNodeStatePositionChangedNotification(unittest.TestCase):
    """Test class for FrameNodeStatePositionChangedNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = (
        b"\x00\x17\x02\x11\x05\x05\xc8\x00\xc8\x00\xf7\xff\xf7\xff"
        b"\xf7\xff\xf7\xff\x00\x00L\xcf\x00\x00\x87"
    )

    def test_bytes(self) -> None:
        """Test FrameNodeStatePositionChangedNotification."""
        frame = FrameNodeStatePositionChangedNotification()
        frame.node_id = 5
        frame.state = OperatingState.DONE
        frame.current_position = Position(position=51200)
        frame.target = Position(position=51200)
        frame.current_position_fp1 = Position()
        frame.current_position_fp2 = Position()
        frame.current_position_fp3 = Position()
        frame.current_position_fp4 = Position()
        frame.remaining_time = 0
        frame.timestamp = 1288634368
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameNodeStatePositionChangedNotification from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameNodeStatePositionChangedNotification))
        self.assertEqual(frame.node_id, 5)
        self.assertEqual(frame.state, OperatingState.DONE)
        self.assertEqual(Position(frame.current_position).position, 51200)
        self.assertEqual(Position(frame.target).position, 51200)
        self.assertEqual(Position(frame.current_position_fp1).position, 63487)
        self.assertEqual(Position(frame.current_position_fp2).position, 63487)
        self.assertEqual(Position(frame.current_position_fp3).position, 63487)
        self.assertEqual(Position(frame.current_position_fp4).position, 63487)
        self.assertEqual(frame.remaining_time, 0)
        self.assertEqual(frame.timestamp, 1288634368)

    def test_str(self) -> None:
        """Test string representation of FrameNodeStatePositionChangedNotification."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        test_ts = datetime.fromtimestamp(1288634368).strftime("%Y-%m-%d %H:%M:%S")
        self.assertEqual(
            str(frame),
            '<FrameNodeStatePositionChangedNotification node_id="5" state="DONE" '
            'current_position="100 %" target="100 %" current_position_fp1="UNKNOWN" '
            'current_position_fp2="UNKNOWN" current_position_fp3="UNKNOWN" '
            'current_position_fp4="UNKNOWN" remaining_time="0" time="{}"/>'.format(
                test_ts
            ),
        )



"""Unit tests for PyVLX PasswordChangeRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FramePasswordChangeRequest
from pyvlx.exception import PyVLXException


class TestFramePasswordChange(unittest.TestCase):
    """Test class for FramePasswordChangeRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FramePasswordChangeRequest."""
        frame = FramePasswordChangeRequest(currentpassword="fnord", newpassword="bfeld")
        self.assertEqual(
            bytes(frame),
            b"\x00C0\x02fnord\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00bfeld\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00i",
        )

    def test_bytes_long_newpw(self):
        """Test FramePasswordChangeRequest with long new password."""
        frame = FramePasswordChangeRequest(currentpassword="fnord", newpassword="x" * 32)
        self.assertEqual(
            bytes(frame), b"\x00C0\x02fnord\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx\x00"
        )

    def test_bytes_long_oldpw(self):
        """Test FramePasswordChangeRequest with long old password."""
        frame = FramePasswordChangeRequest(currentpassword="x" * 32, newpassword="bfeld")
        self.assertEqual(
            bytes(frame), b"\x00C0\x02xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxbfeld\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x18"
        )

    def test_bytes_long_bothpw(self):
        """Test FramePasswordChangeRequest with long passwords."""
        frame = FramePasswordChangeRequest(currentpassword="x" * 32, newpassword="y" * 32)
        self.assertEqual(
            bytes(frame), b"\x00C0\x02xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyq"
        )

    def test_frame_from_raw(self):
        """Test parsing FramePasswordChangeRequest from raw bytes."""
        frame = frame_from_raw(
            b"\x00C0\x02fnord\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00bfeld\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            + b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00i"
        )
        self.assertTrue(isinstance(frame, FramePasswordChangeRequest))
        self.assertEqual(frame.newpassword, "bfeld")
        self.assertEqual(frame.currentpassword, "fnord")

    def test_errors(self):
        """Test FramePasswordChangeRequest with wrong password."""
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeRequest())
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeRequest(currentpassword="fnord"))
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeRequest(currentpassword="fnord", newpassword="x" * 33))
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeRequest(newpassword="fnord", currentpassword="x" * 33))
        with self.assertRaises(PyVLXException):
            bytes(FramePasswordChangeRequest(newpassword="x" * 33, currentpassword="x" * 33))

    def test_str(self):
        """Test string representation of FramePasswordChangeRequest."""
        frame = FramePasswordChangeRequest(currentpassword="fnord", newpassword="bfeld")
        self.assertEqual(str(frame), '<FramePasswordChangeRequest currentpassword="fn****" newpassword="bf****"/>')

    def test_str_no_password(self):
        """Test string representation of FramePasswordChangeRequest with no password."""
        frame = FramePasswordChangeRequest()
        self.assertEqual(str(frame), '<FramePasswordChangeRequest currentpassword="None" newpassword="None"/>')



"""Unit tests for FrameCommandRunStatusNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameCommandRunStatusNotification
from pyvlx.const import RunStatus, StatusReply


class TestFrameCommandRunStatusNotification(unittest.TestCase):
    """Test class FrameCommandRunStatusNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAMES = [
        (
            "execution_completed",
            b"\x00\x10\x03\x02\x03\xe8\x07\x17*\x059\x00\xee\x00\x00\x00\x00\x12",
            RunStatus.EXECUTION_COMPLETED,
            StatusReply.LIMITATION_BY_EMERGENCY,
        ),
        (
            "execution_active",
            b"\x00\x10\x03\x02\x03\xe8\x07\x17*\x059\x02\xe2\x00\x00\x00\x00\x1c",
            RunStatus.EXECUTION_ACTIVE,
            StatusReply.LIMITATION_BY_USER,
        ),
        (
            "execution_failed",
            b"\x00\x10\x03\x02\x03\xe8\x07\x17*\x059\x01\xe3\x00\x00\x00\x00\x1e",
            RunStatus.EXECUTION_FAILED,
            StatusReply.LIMITATION_BY_RAIN,
        ),
    ]

    def test_bytes(self):
        """Test FrameCommandRunStatusNotification."""
        for name, raw, run_status, status_reply in self.EXAMPLE_FRAMES:
            with self.subTest(name=name):
                frame = FrameCommandRunStatusNotification(
                    session_id=1000,
                    status_id=7,
                    index_id=23,
                    node_parameter=42,
                    parameter_value=1337,
                    run_status=run_status,
                    status_reply=status_reply,
                )
                self.assertEqual(bytes(frame), raw)

    def test_frame_from_raw(self):
        """Test parse FrameCommandRunStatusNotification from raw."""
        for name, raw, run_status, status_reply in self.EXAMPLE_FRAMES:
            with self.subTest(name=name):
                frame = frame_from_raw(raw)
                self.assertIsInstance(frame, FrameCommandRunStatusNotification)
                self.assertEqual(frame.session_id, 1000)
                self.assertEqual(frame.status_id, 7)
                self.assertEqual(frame.index_id, 23)
                self.assertEqual(frame.node_parameter, 42)
                self.assertEqual(frame.parameter_value, 1337)
                self.assertEqual(frame.run_status, run_status)
                self.assertEqual(frame.status_reply, status_reply)

    def test_str(self):
        """Test string representation of FrameCommandRunStatusNotification."""
        frame = FrameCommandRunStatusNotification(
            session_id=1000,
            status_id=7,
            index_id=23,
            node_parameter=42,
            parameter_value=1337,
            run_status=RunStatus.EXECUTION_FAILED,
            status_reply=StatusReply.LIMITATION_BY_RAIN,
        )
        self.assertEqual(
            str(frame),
            '<FrameCommandRunStatusNotification session_id="1000" status_id="7" index_id="23" node_parameter="42" '
            'parameter_value="1337" run_status="RunStatus.EXECUTION_FAILED" '
            'status_reply="StatusReply.LIMITATION_BY_RAIN"/>',
        )



"""Unit tests for FrameActivateSceneConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    ActivateSceneConfirmationStatus, FrameActivateSceneConfirmation)


class TestFrameActivateSceneConfirmation(unittest.TestCase):
    """Test class FrameActivateSceneConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameActivateSceneConfirmation."""
        frame = FrameActivateSceneConfirmation(
            session_id=1000, status=ActivateSceneConfirmationStatus.ACCEPTED
        )
        self.assertEqual(bytes(frame), b"\x00\x06\x04\x13\x00\x03\xe8\xfa")

    def test_bytes_error(self):
        """Test failed FrameActivateSceneConfirmation."""
        frame = FrameActivateSceneConfirmation(
            session_id=1000,
            status=ActivateSceneConfirmationStatus.ERROR_REQUEST_REJECTED,
        )
        self.assertEqual(bytes(frame), b"\x00\x06\x04\x13\x02\x03\xe8\xf8")

    def test_frame_from_raw(self):
        """Test parse FrameActivateSceneConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x06\x04\x13\x02\x03\xe8\xf8")
        self.assertTrue(isinstance(frame, FrameActivateSceneConfirmation))
        self.assertEqual(
            frame.status, ActivateSceneConfirmationStatus.ERROR_REQUEST_REJECTED
        )
        self.assertEqual(frame.session_id, 1000)

    def test_str(self):
        """Test string representation of FrameActivateSceneConfirmation."""
        frame = FrameActivateSceneConfirmation(
            session_id=1000, status=ActivateSceneConfirmationStatus.ACCEPTED
        )
        self.assertEqual(
            str(frame),
            '<FrameActivateSceneConfirmation session_id="1000" status="ActivateSceneConfirmationStatus.ACCEPTED"/>',
        )



"""Module for sending get state requests to API in regular periods."""
import asyncio
from typing import TYPE_CHECKING, Any

from .api import GetState
from .api.status_request import StatusRequest
from .exception import PyVLXException
from .log import PYVLXLOG
from .opening_device import Blind, DualRollerShutter

if TYPE_CHECKING:
    from pyvlx import PyVLX


class Heartbeat:
    """Class for sending heartbeats to API."""

    def __init__(
        self, pyvlx: "PyVLX", interval: int = 30, load_all_states: bool = True
    ):
        """Initialize Heartbeat object."""
        PYVLXLOG.debug("Heartbeat __init__")
        self.pyvlx = pyvlx
        self.interval = interval
        self.load_all_states = load_all_states
        self.task: Any = None

    async def _run(self) -> None:
        PYVLXLOG.debug("Heartbeat: task started")
        while True:
            PYVLXLOG.debug("Heartbeat: sleeping")
            await asyncio.sleep(self.interval)
            PYVLXLOG.debug("Heartbeat: pulsing")
            try:
                await self.pulse()
            except (OSError, PyVLXException) as e:
                PYVLXLOG.debug("Heartbeat: pulsing failed: %s", e)

    async def _start(self) -> None:
        if self.task is not None:
            await self.stop()
        PYVLXLOG.debug("Heartbeat: creating task")
        self.task = asyncio.create_task(self._run())

    def start(self) -> None:
        """Start heartbeat."""
        PYVLXLOG.debug("Heartbeat start")
        asyncio.run_coroutine_threadsafe(self._start(), self.pyvlx.loop)

    @property
    def stopped(self) -> bool:
        """Return Heartbeat running state."""
        return self.task is None

    async def stop(self) -> None:
        """Stop heartbeat."""
        if self.task is not None:
            self.task.cancel()
            self.task = None
            PYVLXLOG.debug("Heartbeat stopped")
        else:
            PYVLXLOG.debug("Heartbeat was not running")

    async def pulse(self) -> None:
        """Send get state request to API to keep the connection alive."""
        PYVLXLOG.debug("Heartbeat pulse")
        get_state = GetState(pyvlx=self.pyvlx)
        await get_state.do_api_call()
        if not get_state.success:
            raise PyVLXException("Unable to send get state.")

        # If nodes contain Blind or DualRollerShutter device, refresh orientation or upper/lower curtain positions because House Monitoring
        # delivers wrong values for FP1, FP2 and FP3 parameter
        for node in self.pyvlx.nodes:
            if isinstance(node, (Blind, DualRollerShutter)) or self.load_all_states:
                status_request = StatusRequest(self.pyvlx, node.node_id)
                await status_request.do_api_call()
                # give user requests a chance
                await asyncio.sleep(0.5)



"""Unit tests for FrameGetNetworkSetupRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetNetworkSetupRequest


class TestFrameGetNetworkSetupRequest(unittest.TestCase):
    """Test class for FrameGetNetworkSetupRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetNetworkSetupRequest with NO_TYPE."""
        frame = FrameGetNetworkSetupRequest()
        self.assertEqual(bytes(frame), b"\x00\x03\x00\xe0\xe3")

    def test_frame_from_raw(self):
        """Test parse FrameGetNetworkSetupRequest from raw."""
        frame = frame_from_raw(b"\x00\x03\x00\xe0\xe3")
        self.assertTrue(isinstance(frame, FrameGetNetworkSetupRequest))

    def test_str(self):
        """Test string representation of FrameGetNetworkSetupRequest."""
        frame = FrameGetNetworkSetupRequest()
        self.assertEqual(str(frame), "<FrameGetNetworkSetupRequest/>")



"""Unit tests for FrameSetNodeNameConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FrameSetNodeNameConfirmation, SetNodeNameConfirmationStatus)


class TestFrameSetNodeNameConfirmation(unittest.TestCase):
    """Test class for FrameSetNodeNameConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x05\x02\t\x00\x17\x19"
    EXAMPLE_FRAME_ERR = b"\x00\x05\x02\t\x01\x17\x18"

    def test_bytes(self):
        """Test FrameSetNodeNameConfirmation."""
        frame = FrameSetNodeNameConfirmation(node_id=23)
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_bytes_error(self):
        """Test failed FrameSetNodeNameConfirmation."""
        frame = FrameSetNodeNameConfirmation(
            node_id=23, status=SetNodeNameConfirmationStatus.ERROR_REQUEST_REJECTED
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME_ERR)

    def test_frame_from_raw(self):
        """Test parse FrameSetNodeNameConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameSetNodeNameConfirmation))
        self.assertEqual(frame.status, SetNodeNameConfirmationStatus.OK)
        self.assertEqual(frame.node_id, 23)

    def test_str(self):
        """Test string representation of FrameSetNodeNameConfirmation."""
        frame = FrameSetNodeNameConfirmation(node_id=23)
        self.assertEqual(
            str(frame),
            '<FrameSetNodeNameConfirmation node_id="23" status="SetNodeNameConfirmationStatus.OK"/>',
        )



"""Unit test for roller shutter."""
import unittest

from pyvlx.exception import PyVLXException


# pylint: disable=too-many-public-methods,invalid-name
class TestException(unittest.TestCase):
    """Test class for roller shutter."""

    def test_str(self):
        """Test string representation of PyVLXException."""
        exception = PyVLXException("fnord fnord")
        self.assertEqual(str(exception), '<PyVLXException description="fnord fnord" />')

    def test_str_with_parameter(self):
        """Test string representation of PyVLXException with parameter."""
        exception = PyVLXException("fnord fnord", fnord="fnord", bla="blub")
        self.assertEqual(
            str(exception),
            '<PyVLXException description="fnord fnord" bla="blub" fnord="fnord"/>',
        )



"""Module for retrieving scene list from API."""
from typing import TYPE_CHECKING, Any, Optional

from ..exception import PyVLXException
from ..parameter import Parameter
from .api_event import ApiEvent
from .frames import (
    CommandSendConfirmationStatus, FrameBase,
    FrameCommandRemainingTimeNotification, FrameCommandRunStatusNotification,
    FrameCommandSendConfirmation, FrameCommandSendRequest,
    FrameSessionFinishedNotification)
from .session_id import get_new_session_id

if TYPE_CHECKING:
    from pyvlx import PyVLX


class CommandSend(ApiEvent):
    """Class for sending command to API."""

    def __init__(
            self,
            pyvlx: "PyVLX",
            node_id: int,
            parameter: Parameter,
            active_parameter: int = 0,
            wait_for_completion: bool = True,
            timeout_in_seconds: int = 2,
            **functional_parameter: Any
    ):
        """Initialize SceneList class."""
        super().__init__(pyvlx=pyvlx, timeout_in_seconds=timeout_in_seconds)
        self.success = False
        self.node_id = node_id
        self.parameter = parameter
        self.active_parameter = active_parameter
        self.functional_parameter = functional_parameter
        self.wait_for_completion = wait_for_completion
        self.session_id: Optional[int] = None

    async def handle_frame(self, frame: FrameBase) -> bool:
        """Handle incoming API frame, return True if this was the expected frame."""
        if (
                isinstance(frame, FrameCommandSendConfirmation)
                and frame.session_id == self.session_id
        ):
            if frame.status == CommandSendConfirmationStatus.ACCEPTED:
                self.success = True
            return not self.wait_for_completion
        if (
                isinstance(frame, FrameCommandRemainingTimeNotification)
                and frame.session_id == self.session_id
        ):
            # Ignoring FrameCommandRemainingTimeNotification
            return False
        if (
                isinstance(frame, FrameCommandRunStatusNotification)
                and frame.session_id == self.session_id
        ):
            # At the moment I don't reall understand what the FrameCommandRunStatusNotification is good for.
            # Ignoring these packets for now
            return False
        if (
                isinstance(frame, FrameSessionFinishedNotification)
                and frame.session_id == self.session_id
        ):
            return True
        return False

    async def send(self) -> None:
        """Send frame to KLF200."""
        await self.do_api_call()
        if not self.success:
            raise PyVLXException("Unable to send command")

    def request_frame(self) -> FrameCommandSendRequest:
        """Construct initiating frame."""
        self.session_id = get_new_session_id()
        return FrameCommandSendRequest(
            node_ids=[self.node_id],
            parameter=self.parameter,
            active_parameter=self.active_parameter,
            session_id=self.session_id,
            **self.functional_parameter
        )



"""Unit tests for FrameCommandRemainingTimeNotification."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameCommandRemainingTimeNotification


class TestFrameCommandRemainingTimeNotification(unittest.TestCase):
    """Test class FrameCommandRemainingTimeNotification."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameCommandRemainingTimeNotification."""
        frame = FrameCommandRemainingTimeNotification(
            session_id=1000, index_id=23, node_parameter=42, seconds=1337
        )
        self.assertEqual(bytes(frame), b"\x00\t\x03\x03\x03\xe8\x17*\x059\xe3")

    def test_frame_from_raw(self):
        """Test parse FrameCommandRemainingTimeNotification from raw."""
        frame = frame_from_raw(b"\x00\t\x03\x03\x03\xe8\x17*\x059\xe3")
        self.assertTrue(isinstance(frame, FrameCommandRemainingTimeNotification))
        self.assertEqual(frame.session_id, 1000)
        self.assertEqual(frame.index_id, 23)
        self.assertEqual(frame.node_parameter, 42)
        self.assertEqual(frame.seconds, 1337)

    def test_str(self):
        """Test string representation of FrameCommandRemainingTimeNotification."""
        frame = FrameCommandRemainingTimeNotification(
            session_id=1000, index_id=23, node_parameter=42, seconds=1337
        )
        self.assertEqual(
            str(frame),
            '<FrameCommandRemainingTimeNotification session_id="1000" index_id="23" node_parameter="42" seconds="1337"/>',
        )



"""
Module for PyVLX object.

PyVLX is an asynchronous library for connecting to
a VELUX KLF 200 device for controlling window openers
and roller shutters.
"""
import asyncio
from typing import Optional

from .api import get_limitation
from .api.frames import FrameBase
from .config import Config
from .connection import Connection
from .exception import PyVLXException
from .heartbeat import Heartbeat
from .klf200gateway import Klf200Gateway
from .log import PYVLXLOG
from .node_updater import NodeUpdater
from .nodes import Nodes
from .scenes import Scenes


class PyVLX:
    """Class for PyVLX."""

    def __init__(
        self,
        path: Optional[str] = None,
        host: Optional[str] = None,
        password: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        heartbeat_interval: int = 30,
        heartbeat_load_all_states: bool = True,
    ):
        """Initialize PyVLX class."""
        self.loop = loop or asyncio.get_event_loop()
        self.config = Config(self, path, host, password)
        self.connection = Connection(loop=self.loop, config=self.config)
        self.heartbeat = Heartbeat(
            pyvlx=self,
            interval=heartbeat_interval,
            load_all_states=heartbeat_load_all_states,
        )
        self.node_updater = NodeUpdater(pyvlx=self)
        self.nodes = Nodes(self)
        self.connection.register_frame_received_cb(self.node_updater.process_frame)

        self.scenes = Scenes(self)
        self.version = None
        self.protocol_version = None
        self.klf200 = Klf200Gateway(pyvlx=self)
        self.api_call_semaphore = asyncio.Semaphore(1)  # Limit parallel commands

    async def connect(self) -> None:
        """Connect to KLF 200."""
        PYVLXLOG.debug("Connecting to KLF 200")
        await self.connection.connect()
        assert self.config.password is not None
        await self.klf200.password_enter(password=self.config.password)
        await self.klf200.get_version()
        await self.klf200.get_protocol_version()
        PYVLXLOG.debug(
            "Connected to: %s,  %s",
            str(self.klf200.version),
            str(self.klf200.protocol_version),
        )
        await self.klf200.house_status_monitor_disable(pyvlx=self)
        await self.klf200.get_state()
        await self.klf200.set_utc()
        await self.klf200.get_network_setup()
        await self.klf200.house_status_monitor_enable(pyvlx=self)
        self.heartbeat.start()

    async def reboot_gateway(self) -> None:
        """For Compatibility: Reboot the KLF 200."""
        if not self.get_connected():
            PYVLXLOG.warning("KLF 200 reboot initiated, but gateway is not connected")
        else:
            PYVLXLOG.warning("KLF 200 reboot initiated")
            await self.klf200.reboot()

    def get_connected(self) -> bool:
        """Return whether the gateway is currently connected."""
        return self.connection.connected

    async def check_connected(self) -> None:
        """Check we're connected, and if not, connect."""
        if not self.connection.connected:
            await self.connect()

    async def send_frame(self, frame: FrameBase) -> None:
        """Send frame to API via connection."""
        await self.check_connected()
        self.connection.write(frame)

    async def disconnect(self) -> None:
        """Disconnect from KLF 200."""
        await self.heartbeat.stop()
        if self.connection.connected:
            try:
                # If the connection will be closed while house status monitor is enabled, a reconnection will fail on SSL handshake.
                if self.klf200.house_status_monitor_enabled:
                    await self.klf200.house_status_monitor_disable(pyvlx=self, timeout=5)
                # Reboot KLF200 when disconnecting to avoid unresponsive KLF200.
                await self.klf200.reboot()
            except (OSError, PyVLXException):
                PYVLXLOG.exception("Error during disconnect preparations")
            self.connection.disconnect()
            if self.connection.tasks:
                await asyncio.gather(*self.connection.tasks)  # Wait for all tasks to finish

    async def load_nodes(self, node_id: Optional[int] = None) -> None:
        """Load devices from KLF 200, if no node_id is specified all nodes are loaded."""
        await self.nodes.load(node_id)

    async def load_scenes(self) -> None:
        """Load scenes from KLF 200."""
        await self.scenes.load()

    async def get_limitation(self, node_id: int) -> None:
        """Return limitation."""
        limit = get_limitation.GetLimitation(self, node_id)
        await limit.do_api_call()



"""Module to discover Velux KLF200 devices on the network."""
import asyncio
from asyncio import Event, Future, Task
from dataclasses import dataclass
from typing import Any, Optional, Set

from zeroconf import IPVersion
from zeroconf.asyncio import (
    AsyncServiceBrowser, AsyncServiceInfo, AsyncZeroconf)

SERVICE_STARTS_WITH: str = "VELUX_KLF_LAN"
SERVICE_TYPE: str = "_http._tcp.local."


@dataclass
class VeluxHost():
    """Class to store Velux KLF200 host information."""

    hostname: str
    ip_address: str


def sanitize_hostname(hostname: str) -> str:
    """Sanitize KLF200 hostname."""
    hostname = hostname.upper()
    hostname = hostname.replace("._HTTP._TCP.LOCAL.", "")
    # KLF report it's hostname within Zeroconf ServiceInfo with the "LAN_" prefix, on DHCP requests without
    hostname = hostname.replace("LAN_", "")
    return hostname


class VeluxDiscovery():
    """Class to discover Velux KLF200 devices on the network."""

    hosts: list[VeluxHost] = []
    infos: list[AsyncServiceInfo | None] = []

    def __init__(self, zeroconf: AsyncZeroconf,) -> None:
        """Initialize VeluxDiscovery object."""
        self.zc: AsyncZeroconf = zeroconf

    async def _async_discover_hosts(self, min_wait_time: float, expected_hosts: int | None) -> None:
        """Listen for zeroconf ServiceInfo."""
        self.hosts.clear()
        service_names: list[str] = []
        tasks: Set[Task] = set()
        got_host: Event = Event()

        def add_info_and_host(fut: Future) -> None:
            info: AsyncServiceInfo = fut.result()
            self.infos.append(info)
            hostname = sanitize_hostname(info.name)
            host = VeluxHost(
                hostname=hostname,
                ip_address=info.parsed_addresses(version=IPVersion.V4Only)[0],
            )
            self.hosts.append(host)
            got_host.set()

        def handler(name: str, **kwargs: Any) -> None:  # pylint: disable=W0613:unused-argument
            if name.startswith(SERVICE_STARTS_WITH):
                if name not in service_names:
                    service_names.append(name)
                    task = asyncio.create_task(self.zc.async_get_service_info(type_=SERVICE_TYPE, name=name))
                    tasks.add(task)
                    task.add_done_callback(add_info_and_host)
                    task.add_done_callback(tasks.remove)

        browser: AsyncServiceBrowser = AsyncServiceBrowser(self.zc.zeroconf, SERVICE_TYPE, handlers=[handler])
        if expected_hosts:
            while len(self.hosts) < expected_hosts:
                await got_host.wait()
                got_host.clear()
        while not self.hosts:
            await asyncio.sleep(min_wait_time)
        await browser.async_cancel()
        if tasks:
            await asyncio.gather(*tasks)

    async def async_discover_hosts(
        self,
        timeout: float = 10,
        min_wait_time: float = 3,
        expected_hosts: Optional[int] = None
    ) -> bool:
        """Return true if Velux KLF200 devices found on the network.

        This function creates a zeroconf AsyncServiceBrowser and
        waits min_wait_time (seconds) for ServiceInfos from hosts.
        Some devices may take some time to respond (i.e. if they currently have a high CPU load).
        If one or more Hosts are found, the function cancels the ServiceBrowser and returns true.
        If expected_hosts is set, the function ignores min_wait_time and returns true once expected_hosts are found.
        If timeout (seconds) is exceeded, the function returns false.
        """
        try:
            async with asyncio.timeout(timeout):
                await self._async_discover_hosts(min_wait_time, expected_hosts)
        except TimeoutError:
            return False
        return True



"""Module for handling the TCP connection with Gateway."""
import asyncio
import ssl
import sys
from typing import Callable, Coroutine, List, Optional, Set

from .api.frame_creation import frame_from_raw
from .api.frames import FrameBase
from .config import Config
from .exception import PyVLXException
from .log import PYVLXLOG
from .slip import get_next_slip, is_slip, slip_pack


class SlipTokenizer:
    """Helper class for splitting up binary stream to slip packets."""

    def __init__(self) -> None:
        """Init Tokenizer."""
        self.data = bytes()

    def feed(self, chunk: bytes) -> None:
        """Feed chunk to tokenizer."""
        if not chunk:
            return
        self.data += chunk

    def has_tokens(self) -> bool:
        """Return True if Tokenizer has tokens."""
        return is_slip(self.data)

    def get_next_token(self) -> Optional[bytes]:
        """Get next token from Tokenizer."""
        slip, self.data = get_next_slip(self.data)
        return slip


class TCPTransport(asyncio.Protocol):
    """Class for handling asyncio connection transport."""

    def __init__(
        self,
        frame_received_cb: Callable[[FrameBase], None],
        connection_lost_cb: Callable[[], None],
    ):
        """Init TCPTransport."""
        self.frame_received_cb = frame_received_cb
        self.connection_lost_cb = connection_lost_cb
        self.tokenizer = SlipTokenizer()

    def connection_made(self, transport: object) -> None:
        """Handle sucessful connection."""
        PYVLXLOG.debug("Socket connection to KLF 200 opened")

    def data_received(self, data: bytes) -> None:
        """Handle data received."""
        self.tokenizer.feed(data)
        while self.tokenizer.has_tokens():
            raw = self.tokenizer.get_next_token()
            assert raw is not None

            try:
                frame = frame_from_raw(raw)
                if frame is not None:
                    self.frame_received_cb(frame)
            except PyVLXException:
                PYVLXLOG.error("Error in data_received", exc_info=sys.exc_info())

    def connection_lost(self, exc: object) -> None:
        """Handle lost connection."""
        PYVLXLOG.debug("Socket connection to KLF 200 has been lost")
        self.connection_lost_cb()


CallbackType = Callable[[FrameBase], Coroutine]


class Connection:
    """Class for handling TCP connection."""

    def __init__(self, loop: asyncio.AbstractEventLoop, config: Config):
        """Init TCP connection."""
        self.loop = loop
        self.config = config
        self.transport: Optional[asyncio.Transport] = None
        self.frame_received_cbs: List[CallbackType] = []
        self.connection_closed_cbs: List[Callable[[], Coroutine]] = []
        self.connection_opened_cbs: List[Callable[[], Coroutine]] = []
        self.connected = False
        self.connection_counter = 0
        self.tasks: Set[asyncio.Task] = set()

    def __del__(self) -> None:
        """Destruct connection."""
        self.disconnect()

    def disconnect(self) -> None:
        """Disconnect connection."""
        if self.transport is not None:
            self.transport.close()
            self.transport = None
        self.connected = False
        PYVLXLOG.debug("TCP transport closed.")
        for connection_closed_cb in self.connection_closed_cbs:
            if asyncio.iscoroutine(coroutine := connection_closed_cb()):
                task = self.loop.create_task(coroutine)
                self.tasks.add(task)
                task.add_done_callback(self.tasks.remove)

    async def connect(self) -> None:
        """Connect to gateway via SSL."""
        tcp_client = TCPTransport(self.frame_received_cb, connection_lost_cb=self.on_connection_lost)
        assert self.config.host is not None
        self.transport, _ = await self.loop.create_connection(
            lambda: tcp_client,
            host=self.config.host,
            port=self.config.port,
            ssl=self.create_ssl_context(),
        )
        self.connected = True
        self.connection_counter += 1
        PYVLXLOG.debug(
            "Number of connections since last HA start: %s", self.connection_counter
        )
        for connection_opened_cb in self.connection_opened_cbs:
            if asyncio.iscoroutine(coroutine := connection_opened_cb()):
                task = self.loop.create_task(coroutine)
                self.tasks.add(task)
                task.add_done_callback(self.tasks.remove)

    def register_frame_received_cb(self, callback: CallbackType) -> None:
        """Register frame received callback."""
        self.frame_received_cbs.append(callback)

    def unregister_frame_received_cb(self, callback: CallbackType) -> None:
        """Unregister frame received callback."""
        self.frame_received_cbs.remove(callback)

    def register_connection_closed_cb(self, callback: Callable[[], Coroutine]) -> None:
        """Register connection closed callback."""
        self.connection_closed_cbs.append(callback)

    def unregister_connection_closed_cb(self, callback: Callable[[], Coroutine]) -> None:
        """Unregister connection closed callback."""
        self.connection_closed_cbs.remove(callback)

    def register_connection_opened_cb(self, callback: Callable[[], Coroutine]) -> None:
        """Register connection opened callback."""
        self.connection_opened_cbs.append(callback)

    def unregister_connection_opened_cb(self, callback: Callable[[], Coroutine]) -> None:
        """Unregister connection opened callback."""
        self.connection_opened_cbs.remove(callback)

    def write(self, frame: FrameBase) -> None:
        """Write frame to Bus."""
        if not isinstance(frame, FrameBase):
            raise PyVLXException("Frame not of type FrameBase", *type(frame))
        PYVLXLOG.debug("SEND: %s", frame)
        assert self.transport is not None
        self.transport.write(slip_pack(bytes(frame)))

    @staticmethod
    def create_ssl_context() -> ssl.SSLContext:
        """Create and return SSL Context."""
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        return ssl_context

    def frame_received_cb(self, frame: FrameBase) -> None:
        """Received message."""
        PYVLXLOG.debug("REC: %s", frame)
        for frame_received_cb in self.frame_received_cbs:
            # pylint: disable=not-callable
            task = self.loop.create_task(frame_received_cb(frame))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.remove)

    def on_connection_lost(self) -> None:
        """Server closed connection."""
        self.disconnect()



"""Module for generating a unique session_id."""


_last_session_id = 0
MAX_SESSION_ID = 65535


def get_new_session_id() -> int:
    """Generate new session_id."""
    global _last_session_id  # pylint: disable=global-statement
    _last_session_id = _last_session_id + 1
    if _last_session_id > MAX_SESSION_ID:
        _last_session_id = 1
    return _last_session_id


def set_session_id(session_id: int) -> None:
    """Set session id to value."""
    global _last_session_id  # pylint: disable=global-statement
    _last_session_id = session_id



"""Unit tests for FrameGetLimitationStatusConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames.frame_get_limitation import (
    FrameGetLimitationStatusConfirmation)


class TestFrameGetLimitationStatusConfirmation(unittest.TestCase):
    """Test class for FrameGetLimitationStatusConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetLimitationStatusConfirmation bytes."""
        frame = FrameGetLimitationStatusConfirmation(session_id=1, data=1)
        self.assertEqual(bytes(frame), b'\x00\x06\x03\x13\x00\x01\x01\x16')

        frame = FrameGetLimitationStatusConfirmation(session_id=2, data=0)
        self.assertEqual(bytes(frame), b'\x00\x06\x03\x13\x00\x02\x00\x14')

    def test_frame_from_raw(self):
        """Test parse FrameGetLimitationStatusConfirmation from raw."""
        frame = frame_from_raw(b'\x00\x06\x03\x13\x00\x01\x01\x16')
        self.assertTrue(isinstance(frame, FrameGetLimitationStatusConfirmation))
        self.assertEqual(frame.session_id, 1)
        self.assertEqual(frame.data, 1)

        frame = frame_from_raw(b'\x00\x06\x03\x13\x00\x02\x00\x14')
        self.assertTrue(isinstance(frame, FrameGetLimitationStatusConfirmation))
        self.assertEqual(frame.session_id, 2)
        self.assertEqual(frame.data, 0)

    def test_str(self):
        """Test string representation of FrameGetLimitationStatusConfirmation."""
        frame = FrameGetLimitationStatusConfirmation(session_id=1)
        self.assertEqual(str(frame),
                         '<FrameGetLimitationStatusConfirmation session_id="1" status="None"/>')



"""Unit tests for FrameLeaveLearnStateConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameLeaveLearnStateConfirmation


class TestFrameLeaveLearnStateConfirmation(unittest.TestCase):
    """Test class FrameLeaveLearnStateConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x04\x00\x0f\x00\x0b"

    def test_bytes(self) -> None:
        """Test FrameLeaveLearnStateConfirmation."""
        frame = FrameLeaveLearnStateConfirmation()
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self) -> None:
        """Test parse FrameLeaveLearnStateConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameLeaveLearnStateConfirmation))

    def test_str(self) -> None:
        """Test string representation of FrameLeaveLearnStateConfirmation."""
        frame = FrameLeaveLearnStateConfirmation()
        self.assertEqual(str(frame), '<FrameLeaveLearnStateConfirmation status="LeaveLearnStateConfirmationStatus.FAILED"/>')



"""Unit tests for AliasArray module."""
import unittest

from pyvlx.api.frames.alias_array import AliasArray
from pyvlx.exception import PyVLXException


class TestAliasArray(unittest.TestCase):
    """Test class for AliasArray."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_empty(self) -> None:
        """Test empty alias array."""
        alias_array = AliasArray()
        self.assertEqual(bytes(alias_array), bytes(21))
        self.assertEqual(str(alias_array), "")

    def test_empty_raw(self) -> None:
        """Test empty alias array initialized from raw."""
        alias_array = AliasArray(raw=bytes(21))
        self.assertEqual(bytes(alias_array), bytes(21))
        self.assertEqual(str(alias_array), "")

    def test_from_one_element(self) -> None:
        """Test alias_array with one element."""
        alias_array = AliasArray(raw=b"\x01abcd" + bytes(16))
        self.assertEqual(bytes(alias_array), b"\x01abcd" + bytes(16))
        self.assertEqual(str(alias_array), "6162=6364")

    def test_from_four_elements(self) -> None:
        """Test alias_array with one element."""
        alias_array = AliasArray(raw=b"\x05abcd1234efgh5678ijkl")
        self.assertEqual(bytes(alias_array), b"\x05abcd1234efgh5678ijkl")
        self.assertEqual(
            str(alias_array), "6162=6364, 3132=3334, 6566=6768, 3536=3738, 696a=6b6c"
        )

    def test_deserialize_failure(self) -> None:
        """Test error while deserializing."""
        with self.assertRaises(PyVLXException):
            AliasArray(raw="string")  # type: ignore[arg-type]
        with self.assertRaises(PyVLXException):
            AliasArray(raw=bytes(20))
        with self.assertRaises(PyVLXException):
            AliasArray(raw=b"\x06" + bytes(20))



"""Example implementation for PyVLX."""
import asyncio
from pyvlx import PyVLX


async def main():
    """Load devices and scenes, run first scene."""
    pyvlx = PyVLX('pyvlx.yaml')
    # Alternative:
    # pyvlx = PyVLX(host="192.168.2.127", password="velux123", timeout=60)

    await pyvlx.load_devices()
    print(pyvlx.devices[1])
    print(pyvlx.devices['Fenster 4'])

    await pyvlx.load_scenes()
    print(pyvlx.scenes[0])
    print(pyvlx.scenes['Bath Closed'])

    # opening/ closing windows by running scenes, yay!
    await pyvlx.scenes[1].run()

    await pyvlx.disconnect()


# pylint: disable=invalid-name
loop = asyncio.get_event_loop()
loop.run_until_complete(main())
loop.close()



"""Unit tests for FrameHouseStatusMonitorDisableConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameHouseStatusMonitorDisableConfirmation


class TestFrameHouseStatusMonitorDisableConfirmation(unittest.TestCase):
    """Test class FrameHouseStatusMonitorDisableConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x03\x02CB"

    def test_bytes(self):
        """Test FrameHouseStatusMonitorDisableConfirmation."""
        frame = FrameHouseStatusMonitorDisableConfirmation()
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameHouseStatusMonitorDisableConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameHouseStatusMonitorDisableConfirmation))

    def test_str(self):
        """Test string representation of FrameHouseStatusMonitorDisableConfirmation."""
        frame = FrameHouseStatusMonitorDisableConfirmation()
        self.assertEqual(str(frame), "<FrameHouseStatusMonitorDisableConfirmation/>")



"""Module for storing and accessing scene list."""
from typing import TYPE_CHECKING, Iterator, List, Union

from .api import GetSceneList
from .exception import PyVLXException
from .scene import Scene

if TYPE_CHECKING:
    from pyvlx import PyVLX


class Scenes:
    """Class for storing and accessing ."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize Scenes class."""
        self.pyvlx = pyvlx
        self.__scenes: List[Scene] = []

    def __iter__(self) -> Iterator[Scene]:
        """Iterate."""
        yield from self.__scenes

    def __getitem__(self, key: Union[str, int]) -> Scene:
        """Return scene by name or by index."""
        if isinstance(key, int):
            for scene in self.__scenes:
                if scene.scene_id == key:
                    return scene
        for scene in self.__scenes:
            if scene.name == key:
                return scene
        raise KeyError

    def __len__(self) -> int:
        """Return number of scenes."""
        return len(self.__scenes)

    def add(self, scene: Scene) -> None:
        """Add scene, replace existing scene if scene with scene_id is present."""
        if not isinstance(scene, Scene):
            raise TypeError()
        for i, j in enumerate(self.__scenes):
            if j.scene_id == scene.scene_id:
                self.__scenes[i] = scene
                return
        self.__scenes.append(scene)

    def clear(self) -> None:
        """Clear internal scenes array."""
        self.__scenes = []

    async def load(self) -> None:
        """Load scenes from KLF 200."""
        get_scene_list = GetSceneList(pyvlx=self.pyvlx)
        await get_scene_list.do_api_call()
        if not get_scene_list.success:
            raise PyVLXException("Unable to retrieve scene information")
        for scene in get_scene_list.scenes:
            self.add(Scene(pyvlx=self.pyvlx, scene_id=scene[0], name=scene[1]))



"""Unit tests string_helper module."""
import unittest

from pyvlx.exception import PyVLXException
from pyvlx.string_helper import bytes_to_string, string_to_bytes


class TestString(unittest.TestCase):
    """Test class for String encoding/decoding."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_encoding(self) -> None:
        """Test simple encoding."""
        self.assertEqual(string_to_bytes("fnord", 10), b"fnord" + bytes(5))

    def test_encoding_without_padding(self) -> None:
        """Test encoding with exact match of size."""
        self.assertEqual(string_to_bytes("fnord", 5), b"fnord")

    def test_encoding_failure(self) -> None:
        """Test error while encoding."""
        with self.assertRaises(PyVLXException):
            string_to_bytes("fnord", 4)

    def test_decoding(self) -> None:
        """Test decoding of string."""
        self.assertEqual(bytes_to_string(b"fnord" + bytes(5)), "fnord")

    def test_decoding_without_padding(self) -> None:
        """Test decoding of string without padding."""
        self.assertEqual(bytes_to_string(b"fnord"), "fnord")

    def test_encode_utf8(self) -> None:
        """Test encoding a string with special characters."""
        self.assertEqual(
            string_to_bytes("Fenster Büro", 16), b"Fenster B\xc3\xbcro\x00\x00\x00"
        )

    def test_decode_utf8(self) -> None:
        """Test decoding a string with special characters."""
        self.assertEqual(
            bytes_to_string(b"Fenster B\xc3\xbcro\x00\x00\x00"), "Fenster Büro"
        )



"""Unit tests for FrameGetStateConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    FrameGetStateConfirmation, GatewayState, GatewaySubState)


class TestFrameGetStateConfirmation(unittest.TestCase):
    """Test class FrameGetStateConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\t\x00\r\x03\x80\x00\x00\x00\x00\x87"

    def test_bytes(self):
        """Test FrameGetStateConfirmation with NO_TYPE."""
        frame = FrameGetStateConfirmation()
        frame.gateway_state = GatewayState.BEACON_MODE_NOT_CONFIGURED
        frame.gateway_sub_state = GatewaySubState.PERFORMING_TASK_COMMAND
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameGetStateConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameGetStateConfirmation))
        self.assertEqual(frame.gateway_state, GatewayState.BEACON_MODE_NOT_CONFIGURED)
        self.assertEqual(
            frame.gateway_sub_state, GatewaySubState.PERFORMING_TASK_COMMAND
        )

    def test_str(self):
        """Test string representation of FrameGetStateConfirmation."""
        frame = FrameGetStateConfirmation()
        frame.gateway_state = GatewayState.BEACON_MODE_NOT_CONFIGURED
        frame.gateway_sub_state = GatewaySubState.PERFORMING_TASK_COMMAND
        self.assertEqual(
            str(frame),
            "<FrameGetStateConfirmation "
            'gateway_state="GatewayState.BEACON_MODE_NOT_CONFIGURED" '
            'gateway_sub_state="GatewaySubState.PERFORMING_TASK_COMMAND"/>',
        )



"""Module for storing devices."""

import json
from .device import Device
from .window import Window
from .rollershutter import RollerShutter
from .blind import Blind
from .exception import PyVLXException


class Devices:
    """Object for storing devices."""

    def __init__(self, pyvlx):
        """Initialize Devices class."""
        self.pyvlx = pyvlx
        self.__devices = []

    def __iter__(self):
        """Iterator."""
        yield from self.__devices

    def __getitem__(self, key):
        """Return device by name or by index."""
        for device in self.__devices:
            if device.name == key:
                return device
        if isinstance(key, int):
            return self.__devices[key]
        raise KeyError

    def __len__(self):
        """Return number of devices."""
        return len(self.__devices)

    def add(self, device):
        """Add device."""
        if not isinstance(device, Device):
            raise TypeError()
        self.__devices.append(device)

    async def load(self):
        """Load devices from KLF 200."""
        json_response = await self.pyvlx.interface.api_call('products', 'get')
        self.data_import(json_response)

    def data_import(self, json_response):
        """Import data from json response."""
        if 'data' not in json_response:
            raise PyVLXException('no element data found: {0}'.format(
                json.dumps(json_response)))
        data = json_response['data']

        for item in data:
            if 'category' not in item:
                raise PyVLXException('no element category: {0}'.format(
                    json.dumps(item)))
            category = item['category']
            if category == 'Window opener':
                self.load_window_opener(item)
            elif category in ['Roller shutter', 'Dual Shutter']:
                self.load_roller_shutter(item)
            elif category in ['Blind']:
                self.load_blind(item)
            else:
                self.pyvlx.logger.warning(
                    'WARNING: Could not parse product: %s', category)

    def load_window_opener(self, item):
        """Load window opener from JSON."""
        window = Window.from_config(self.pyvlx, item)
        self.add(window)

    def load_roller_shutter(self, item):
        """Load roller shutter from JSON."""
        rollershutter = RollerShutter.from_config(self.pyvlx, item)
        self.add(rollershutter)

    def load_blind(self, item):
        """Load blind from JSON."""
        blind = Blind.from_config(self.pyvlx, item)
        self.add(blind)



"""Unit tests for FrameCommandSendConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import (
    CommandSendConfirmationStatus, FrameCommandSendConfirmation)


class TestFrameCommandSendConfirmation(unittest.TestCase):
    """Test class FrameCommandSendConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameCommandSendConfirmation."""
        frame = FrameCommandSendConfirmation(
            session_id=1000, status=CommandSendConfirmationStatus.ACCEPTED
        )
        self.assertEqual(bytes(frame), b"\x00\x06\x03\x01\x03\xe8\x01\xee")

    def test_bytes_error(self):
        """Test failed FrameCommandSendConfirmation."""
        frame = FrameCommandSendConfirmation(
            session_id=1000, status=CommandSendConfirmationStatus.REJECTED
        )
        self.assertEqual(bytes(frame), b"\x00\x06\x03\x01\x03\xe8\x00\xef")

    def test_frame_from_raw(self):
        """Test parse FrameCommandSendConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x06\x03\x01\x03\xe8\x00\xef")
        self.assertTrue(isinstance(frame, FrameCommandSendConfirmation))
        self.assertEqual(frame.status, CommandSendConfirmationStatus.REJECTED)
        self.assertEqual(frame.session_id, 1000)

    def test_str(self):
        """Test string representation of FrameCommandSendConfirmation."""
        frame = FrameCommandSendConfirmation(
            session_id=1000, status=CommandSendConfirmationStatus.ACCEPTED
        )
        self.assertEqual(
            str(frame),
            '<FrameCommandSendConfirmation session_id="1000" status="CommandSendConfirmationStatus.ACCEPTED"/>',
        )



"""Unit tests for FrameGetSceneListRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetSceneListRequest


class TestFrameGetSceneListRequest(unittest.TestCase):
    """Test class for FrameGetSceneListRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetSceneListRequest with NO_TYPE."""
        frame = FrameGetSceneListRequest()
        self.assertEqual(bytes(frame), b"\x00\x03\x04\x0c\x0b")

    def test_frame_from_raw(self):
        """Test parse FrameGetSceneListRequest from raw."""
        frame = frame_from_raw(b"\x00\x03\x04\x0c\x0b")
        self.assertTrue(isinstance(frame, FrameGetSceneListRequest))

    def test_str(self):
        """Test string representation of FrameGetSceneListRequest."""
        frame = FrameGetSceneListRequest()
        self.assertEqual(str(frame), "<FrameGetSceneListRequest/>")



"""Unit tests for FrameGetVersionConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetVersionConfirmation


class TestFrameGetVersionConfirmation(unittest.TestCase):
    """Test class for FrameGetVersionConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x0c\x00\t\x01\x02\x03\x04\x05\x06\x17\x0e\x03\x18"

    def test_bytes(self):
        """Test FrameGetVersionConfirmation."""
        frame = FrameGetVersionConfirmation(
            software_version="1.2.3.4.5.6", hardware_version=23
        )
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameGetVersionConfirmation from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameGetVersionConfirmation))
        self.assertEqual(frame.software_version, "1.2.3.4.5.6")
        self.assertEqual(frame.hardware_version, 23)
        self.assertEqual(frame.product, "KLF 200")

    def test_str(self):
        """Test string representation of FrameGetVersionConfirmation."""
        frame = FrameGetVersionConfirmation(
            software_version="1.2.3.4.5.6", hardware_version=23
        )
        self.assertEqual(
            str(frame),
            '<FrameGetVersionConfirmation software_version="1.2.3.4.5.6" hardware_version="23" product="KLF 200"/>',
        )

    def test_version(self):
        """Test version string."""
        frame = FrameGetVersionConfirmation(
            software_version="1.2.3.4.5.6", hardware_version=23
        )
        self.assertEqual(
            frame.version,
            "KLF 200: Software version: 1.2.3.4.5.6, hardware version: 23",
        )

    def test_product(self):
        """Test formatting of product."""
        frame = FrameGetVersionConfirmation()
        self.assertEqual(frame.product, "KLF 200")
        frame.product_type = 42
        frame.product_group = 23
        self.assertEqual(frame.product, "Unknown Product: 23:42")



"""PyVLX - controling VELUX windows with Python."""

# flake8: noqa
from .pyvlx import PyVLX
from .window import Window
from .rollershutter import RollerShutter
from .blind import Blind
from .scene import Scene
from .scenes import Scenes
from .device import Device
from .devices import Devices
from .config import Config
from .interface import Interface
from .exception import PyVLXException, InvalidToken



"""Unit tests for FrameGetAllNodesInformationRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameGetAllNodesInformationRequest


class TestFrameGetAllNodesInformationRequest(unittest.TestCase):
    """Test class for FrameGetAllNodesInformationRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameGetAllNodesInformationRequest with NO_TYPE."""
        frame = FrameGetAllNodesInformationRequest()
        self.assertEqual(bytes(frame), b"\x00\x03\x02\x02\x03")

    def test_frame_from_raw(self):
        """Test parse FrameGetAllNodesInformationRequest from raw."""
        frame = frame_from_raw(b"\x00\x03\x02\x02\x03")
        self.assertTrue(isinstance(frame, FrameGetAllNodesInformationRequest))

    def test_str(self):
        """Test string representation of FrameGetAllNodesInformationRequest."""
        frame = FrameGetAllNodesInformationRequest()
        self.assertEqual(str(frame), "<FrameGetAllNodesInformationRequest/>")



"""Helper module for Node objects."""
from typing import TYPE_CHECKING, Optional, Union

from .api.frames import (
    FrameGetAllNodesInformationNotification,
    FrameGetNodeInformationNotification)
from .const import NodeTypeWithSubtype
from .dimmable_device import ExteriorHeating, Light, OnOffLight
from .log import PYVLXLOG
from .node import Node
from .on_off_switch import OnOffSwitch
from .opening_device import (
    Awning, Blade, Blind, DualRollerShutter, GarageDoor, Gate, RollerShutter,
    Window)

if TYPE_CHECKING:
    from pyvlx import PyVLX


def convert_frame_to_node(
    pyvlx: "PyVLX",
    frame: Union[
        FrameGetNodeInformationNotification, FrameGetAllNodesInformationNotification
    ],
) -> Optional[Node]:
    """Convert FrameGet[All]Node[s]InformationNotification into Node object."""
    # pylint: disable=too-many-return-statements

    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER:
        return Window(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
            rain_sensor=False,
        )

    if frame.node_type == NodeTypeWithSubtype.WINDOW_OPENER_WITH_RAIN_SENSOR:
        return Window(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
            rain_sensor=True,
        )

    if frame.node_type == NodeTypeWithSubtype.DUAL_ROLLER_SHUTTER:
        return DualRollerShutter(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type in [
        NodeTypeWithSubtype.ROLLER_SHUTTER,
        NodeTypeWithSubtype.SWINGING_SHUTTERS,
    ]:
        return RollerShutter(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type in [
        NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND,
        NodeTypeWithSubtype.VERTICAL_INTERIOR_BLINDS,
        NodeTypeWithSubtype.INTERIOR_VENETIAN_BLIND,
    ]:
        return RollerShutter(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    # Blinds have position and orientation (inherit frame.current_position_fp3) attribute
    if frame.node_type in [
        NodeTypeWithSubtype.EXTERIOR_VENETIAN_BLIND,
        NodeTypeWithSubtype.ADJUSTABLE_SLATS_ROLLING_SHUTTER,
        NodeTypeWithSubtype.LOUVER_BLIND,
    ]:
        return Blind(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type in [
        NodeTypeWithSubtype.VERTICAL_EXTERIOR_AWNING,
        NodeTypeWithSubtype.HORIZONTAL_AWNING,
        NodeTypeWithSubtype.HORIZONTAL_AWNING_ALT,
    ]:
        return Awning(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.ON_OFF_SWITCH:
        return OnOffSwitch(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    if frame.node_type in [
        NodeTypeWithSubtype.GARAGE_DOOR_OPENER,
        NodeTypeWithSubtype.LINEAR_ANGULAR_POSITION_OF_GARAGE_DOOR,
    ]:
        return GarageDoor(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.GATE_OPENER:
        return Gate(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.GATE_OPENER_ANGULAR_POSITION:
        return Gate(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.BLADE_OPENER:
        return Blade(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
            position_parameter=frame.current_position,
        )

    if frame.node_type == NodeTypeWithSubtype.LIGHT:
        return Light(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    if frame.node_type == NodeTypeWithSubtype.LIGHT_ON_OFF:
        return OnOffLight(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    if frame.node_type == NodeTypeWithSubtype.EXTERIOR_HEATING:
        return ExteriorHeating(
            pyvlx=pyvlx,
            node_id=frame.node_id,
            name=frame.name,
            serial_number=frame.serial_number,
        )

    PYVLXLOG.warning("%s not implemented", frame.node_type)
    return None



"""Unit tests for FrameHouseStatusMonitorEnableRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameHouseStatusMonitorEnableRequest


class TestFrameHouseStatusMonitorEnableRequest(unittest.TestCase):
    """Test class FrameHouseStatusMonitorEnableRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    EXAMPLE_FRAME = b"\x00\x03\x02@A"

    def test_bytes(self):
        """Test FrameHouseStatusMonitorEnableRequest."""
        frame = FrameHouseStatusMonitorEnableRequest()
        self.assertEqual(bytes(frame), self.EXAMPLE_FRAME)

    def test_frame_from_raw(self):
        """Test parse FrameHouseStatusMonitorEnableRequest from raw."""
        frame = frame_from_raw(self.EXAMPLE_FRAME)
        self.assertTrue(isinstance(frame, FrameHouseStatusMonitorEnableRequest))

    def test_str(self):
        """Test string representation of FrameHouseStatusMonitorEnableRequest."""
        frame = FrameHouseStatusMonitorEnableRequest()
        self.assertEqual(str(frame), "<FrameHouseStatusMonitorEnableRequest/>")



"""Unit test for roller shutter."""
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock, patch

from pyvlx import (
    Awning, Blade, Blind, CurrentPosition, OpeningDevice, Parameter, Position,
    PyVLX, RollerShutter, Window)
from pyvlx.connection import Connection
from pyvlx.const import Velocity


# pylint: disable=too-many-public-methods,invalid-name
class TestOpeningDevice(IsolatedAsyncioTestCase):
    """Test class for roller shutter."""

    mocked_pyvlx = MagicMock(spec=PyVLX)
    connection = MagicMock(spec=Connection)
    mocked_pyvlx.attach_mock(mock=connection, attribute="connection")

    @patch("pyvlx.api.CommandSend.send", new_callable=AsyncMock)
    @patch("pyvlx.Node.after_update", new_callable=AsyncMock)
    async def test_set_position(self, commandSend: AsyncMock, afterUpdate: AsyncMock) -> None:
        """Test set_position of OpeningDevice object."""
        opening_device = OpeningDevice(pyvlx=self.mocked_pyvlx, node_id=23, name="Test device")
        await opening_device.set_position(position=Position(position_percent=100))
        assert commandSend.called
        assert afterUpdate.called

    @patch("pyvlx.opening_device.OpeningDevice.set_position", new_callable=AsyncMock)
    async def test_open(self, set_position: AsyncMock) -> None:
        """Test open function of OpeningDevice object."""
        opening_device = OpeningDevice(pyvlx=self.mocked_pyvlx, node_id=23, name="Test device")
        velocity = Velocity.DEFAULT
        wait_for_completion = False
        await opening_device.open(velocity=velocity, wait_for_completion=wait_for_completion)
        set_position.assert_awaited_once_with(
            position=Position(position_percent=opening_device.open_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion)

    @patch("pyvlx.opening_device.OpeningDevice.set_position", new_callable=AsyncMock)
    async def test_close(self, set_position: AsyncMock) -> None:
        """Test close function of OpeningDevice object."""
        opening_device = OpeningDevice(pyvlx=self.mocked_pyvlx, node_id=23, name="Test device")
        velocity = Velocity.DEFAULT
        wait_for_completion = False
        await opening_device.close(velocity=velocity, wait_for_completion=wait_for_completion)
        set_position.assert_awaited_once_with(
            position=Position(position_percent=opening_device.close_position_target),
            velocity=velocity,
            wait_for_completion=wait_for_completion)

    @patch("pyvlx.opening_device.OpeningDevice.set_position", new_callable=AsyncMock)
    async def test_stop(self, set_position: AsyncMock) -> None:
        """Test stop function of OpeningDevice object."""
        opening_device = OpeningDevice(pyvlx=self.mocked_pyvlx, node_id=23, name="Test device")
        wait_for_completion = False
        await opening_device.stop(wait_for_completion=wait_for_completion)
        set_position.assert_awaited_once_with(
            position=CurrentPosition(),
            wait_for_completion=wait_for_completion)

    def test_window_str(self) -> None:
        """Test string representation of Window object."""
        pyvlx = self.mocked_pyvlx
        window = Window(
            pyvlx=pyvlx,
            node_id=23,
            name="Test Window",
            rain_sensor=True,
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(window),
            '<Window name="Test Window" node_id="23" rain_sensor=True serial_number="aa:bb:aa:bb:aa:bb:aa:23" position="UNKNOWN"/>',
        )

    def test_blind_str(self) -> None:
        """Test string representation of Blind object."""
        pyvlx = self.mocked_pyvlx
        blind = Blind(
            pyvlx=pyvlx,
            node_id=23,
            name="Test Blind",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(blind),
            '<Blind name="Test Blind" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23" position="UNKNOWN"/>',
        )

    def test_roller_shutter_str(self) -> None:
        """Test string representation of RolllerShutter object."""
        pyvlx = self.mocked_pyvlx
        roller_shutter = RollerShutter(
            pyvlx=pyvlx,
            node_id=23,
            name="Test Roller Shutter",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
            position_parameter=Parameter(Parameter.from_int(int(0.97 * Parameter.MAX))),
        )
        self.assertEqual(
            str(roller_shutter),
            '<RollerShutter name="Test Roller Shutter" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23" position="97 %"/>',
        )

    def test_blade_str(self) -> None:
        """Test string representation of Blade object."""
        pyvlx = self.mocked_pyvlx
        blade = Blade(
            pyvlx=pyvlx,
            node_id=23,
            name="Test Blade",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(blade),
            '<Blade name="Test Blade" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23" position="UNKNOWN"/>',
        )

    def test_awning_str(self) -> None:
        """Test string representation of Awning object."""
        pyvlx = self.mocked_pyvlx
        awning = Awning(
            pyvlx=pyvlx,
            node_id=23,
            name="Test Awning",
            serial_number="aa:bb:aa:bb:aa:bb:aa:23",
        )
        self.assertEqual(
            str(awning),
            '<Awning name="Test Awning" node_id="23" serial_number="aa:bb:aa:bb:aa:bb:aa:23" position="UNKNOWN"/>',
        )

    def test_eq(self) -> None:
        """Testing eq method with positive results."""
        pyvlx = self.mocked_pyvlx
        node1 = Blind(
            pyvlx=pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        node2 = Blind(
            pyvlx=pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        self.assertEqual(node1, node2)

    def test_nq(self) -> None:
        """Testing eq method with negative results."""
        pyvlx = self.mocked_pyvlx
        node1 = Blind(
            pyvlx=pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        node2 = Blind(
            pyvlx=pyvlx, node_id=24, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:24"
        )
        node3 = RollerShutter(
            pyvlx=pyvlx, node_id=23, name="xxx", serial_number="aa:bb:aa:bb:aa:bb:aa:23"
        )
        self.assertNotEqual(node1, node2)
        self.assertNotEqual(node2, node3)
        self.assertNotEqual(node3, node1)



"""Module for configuration."""
import yaml
from .exception import PyVLXException


class Config:
    """Object for configuration."""

    def __init__(self, pyvlx, path=None, host=None, password=None):
        """Initialize Config class."""
        self.pyvlx = pyvlx
        if path is not None:
            self.read_config(path)
        if host is not None:
            self.host = host
        if password is not None:
            self.password = password

    def read_config(self, path):
        """Read configuration file."""
        self.pyvlx.logger.info('Reading config file: ', path)
        try:
            with open(path, 'r') as filehandle:
                doc = yaml.load(filehandle)
                if 'config' not in doc:
                    raise PyVLXException('no element config found in: {0}'.format(path))
                if 'host' not in doc['config']:
                    raise PyVLXException('no element host found in: {0}'.format(path))
                if 'password' not in doc['config']:
                    raise PyVLXException('no element password found in: {0}'.format(path))
                self.host = doc['config']['host']
                self.password = doc['config']['password']
        except FileNotFoundError as ex:
            raise PyVLXException('file does not exist: {0}'.format(ex))



"""Unit tests for FrameDiscoverNodesRequest."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameDiscoverNodesRequest
from pyvlx.const import NodeType


class TestFrameNodeDiscover(unittest.TestCase):
    """Test class for FrameDiscoverNodesRequest."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameDiscoverNodesRequest with NO_TYPE."""
        frame = FrameDiscoverNodesRequest()
        self.assertEqual(bytes(frame), b"\x00\x04\x01\x03\x00\x06")

    def test_bytes_light(self):
        """Test FrameDiscoverNodesRequest with Light."""
        frame = FrameDiscoverNodesRequest(NodeType.LIGHT)
        self.assertEqual(bytes(frame), b"\x00\x04\x01\x03\x06\x00")

    def test_frame_from_raw(self):
        """Test parse FrameDiscoverNodesRequest from raw."""
        frame = frame_from_raw(b"\x00\x04\x01\x03\x06\x00")
        self.assertTrue(isinstance(frame, FrameDiscoverNodesRequest))
        self.assertEqual(frame.node_type, NodeType.LIGHT)

    def test_str(self):
        """Test string representation of FrameDiscoverNodesRequest."""
        frame = FrameDiscoverNodesRequest(NodeType.LIGHT)
        self.assertEqual(
            str(frame), '<FrameDiscoverNodesRequest node_type="NodeType.LIGHT"/>'
        )



"""Module for storing scenes."""

import json
from .scene import Scene
from .exception import PyVLXException


class Scenes:
    """Object for storing scenes."""

    def __init__(self, pyvlx):
        """Initialize Scenes class."""
        self.pyvlx = pyvlx
        self.__scenes = []

    def __iter__(self):
        """Iterator."""
        yield from self.__scenes

    def __getitem__(self, key):
        """Return scene by name or by index."""
        for scene in self.__scenes:
            if scene.name == key:
                return scene
        if isinstance(key, int):
            return self.__scenes[key]
        raise KeyError

    def __len__(self):
        """Return number of scenes."""
        return len(self.__scenes)

    def add(self, scene):
        """Add scene."""
        if not isinstance(scene, Scene):
            raise TypeError()
        self.__scenes.append(scene)

    async def load(self):
        """Load scenes from KLF 200."""
        json_response = await self.pyvlx.interface.api_call('scenes', 'get')
        self.data_import(json_response)

    def data_import(self, json_response):
        """Import scenes from JSON response."""
        if 'data' not in json_response:
            raise PyVLXException('no element data found: {0}'.format(
                json.dumps(json_response)))
        data = json_response['data']
        for item in data:
            self.load_scene(item)

    def load_scene(self, item):
        """Load scene from json."""
        scene = Scene.from_config(self.pyvlx, item)
        self.add(scene)



"""Unit tests for frame module."""
import unittest

from pyvlx import PyVLXException
from pyvlx.api.frames.frame import FrameBase
from pyvlx.const import Command


class TestFrame(unittest.TestCase):
    """Test class FrameBase class."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_validate_payload_len_no_payload_len_defined(self):
        """Test validate_payload_len_function without any PAYLOAD_LEN defined."""
        frame = FrameBase(command=Command.GW_GET_NODE_INFORMATION_REQ)
        frame.validate_payload_len(bytes(23))

    def test_validate_payload_len_payload_len_defined(self):
        """Test validate_payload_len_function with defined PAYLOAD_LEN."""
        frame = FrameBase(command=Command.GW_GET_NODE_INFORMATION_REQ)
        # pylint: disable=invalid-name
        frame.PAYLOAD_LEN = 23
        frame.validate_payload_len(bytes(23))

    def test_validate_payload_len_payload_len_error(self):
        """Test validate_payload_len_function with wrong PAYLOAD_LEN."""
        frame = FrameBase(command=Command.GW_GET_NODE_INFORMATION_REQ)
        frame.PAYLOAD_LEN = 42
        with self.assertRaises(PyVLXException):
            frame.validate_payload_len(bytes(23))



"""Module for updating nodes via frames."""
import datetime
from typing import TYPE_CHECKING, Any

from .api.frames import (
    FrameBase, FrameGetAllNodesInformationNotification,
    FrameNodeStatePositionChangedNotification, FrameStatusRequestNotification)
from .const import NodeParameter, OperatingState
from .dimmable_device import DimmableDevice
from .log import PYVLXLOG
from .on_off_switch import OnOffSwitch
from .opening_device import Blind, DualRollerShutter, OpeningDevice
from .parameter import Intensity, Parameter, Position, SwitchParameter

if TYPE_CHECKING:
    from pyvlx import PyVLX


class NodeUpdater:
    """Class for updating nodes via incoming frames,  usually received by house monitor."""

    def __init__(self, pyvlx: "PyVLX"):
        """Initialize NodeUpdater object."""
        self.pyvlx = pyvlx

    async def process_frame_status_request_notification(
        self, frame: FrameStatusRequestNotification
    ) -> None:
        """Process FrameStatusRequestNotification."""
        PYVLXLOG.debug("NodeUpdater process frame: %s", frame)
        if frame.node_id not in self.pyvlx.nodes:
            return
        node = self.pyvlx.nodes[frame.node_id]
        if isinstance(node, Blind):
            if NodeParameter(0) not in frame.parameter_data:  # MP missing in frame
                return
            if NodeParameter(3) not in frame.parameter_data:  # FP3 missing in frame
                return
            position = Position(frame.parameter_data[NodeParameter(0)])
            orientation = Position(frame.parameter_data[NodeParameter(3)])
            if position.position <= Parameter.MAX:
                node.position = position
                PYVLXLOG.debug("%s position changed to: %s", node.name, position)
            if orientation.position <= Parameter.MAX:
                node.orientation = orientation
                PYVLXLOG.debug("%s orientation changed to: %s", node.name, orientation)
            await node.after_update()

        if isinstance(node, DualRollerShutter):
            if NodeParameter(0) not in frame.parameter_data:  # MP missing in frame
                return
            if NodeParameter(1) not in frame.parameter_data:  # FP1 missing in frame
                return
            if NodeParameter(2) not in frame.parameter_data:  # FP2 missing in frame
                return
            position = Position(frame.parameter_data[NodeParameter(0)])
            position_upper_curtain = Position(frame.parameter_data[NodeParameter(1)])
            position_lower_curtain = Position(frame.parameter_data[NodeParameter(2)])
            if position.position <= Parameter.MAX:
                node.position = position
                PYVLXLOG.debug("%s position changed to: %s", node.name, position)
            if position_upper_curtain.position <= Parameter.MAX:
                node.position_upper_curtain = position_upper_curtain
                PYVLXLOG.debug(
                    "%s position upper curtain changed to: %s",
                    node.name,
                    position_upper_curtain,
                )
            if position_lower_curtain.position <= Parameter.MAX:
                node.position_lower_curtain = position_lower_curtain
                PYVLXLOG.debug(
                    "%s position lower curtain changed to: %s",
                    node.name,
                    position_lower_curtain,
                )
            await node.after_update()

    async def process_frame(self, frame: FrameBase) -> None:
        """Update nodes via frame, usually received by house monitor."""
        if isinstance(
            frame,
            (
                FrameGetAllNodesInformationNotification,
                FrameNodeStatePositionChangedNotification,
            ),
        ):
            PYVLXLOG.debug("NodeUpdater process frame: %s", frame)
            if frame.node_id not in self.pyvlx.nodes:
                return
            node = self.pyvlx.nodes[frame.node_id]
            position = Position(frame.current_position)
            target: Any = Position(frame.target)
            # KLF transmits for functional parameters basically always 'No feed-back value known’ (0xF7FF).
            # In home assistant this cause unreasonable values like -23%. Therefore a check is implemented
            # whether the frame parameter is inside the maximum range.

            # Set opening device status
            if isinstance(node, OpeningDevice):
                if (position.position > target.position <= Parameter.MAX) and (
                    (frame.state == OperatingState.EXECUTING)
                    or frame.remaining_time > 0
                ):
                    node.is_opening = True
                    PYVLXLOG.debug("%s is opening", node.name)
                    node.state_received_at = datetime.datetime.now()
                    node.estimated_completion = (
                        node.state_received_at
                        + datetime.timedelta(0, frame.remaining_time)
                    )
                    PYVLXLOG.debug(
                        "%s will be opening until", node.estimated_completion
                    )
                elif (position.position < target.position <= Parameter.MAX) and (
                    (frame.state == OperatingState.EXECUTING)
                    or frame.remaining_time > 0
                ):
                    node.is_closing = True
                    PYVLXLOG.debug("%s is closing", node.name)
                    node.state_received_at = datetime.datetime.now()
                    node.estimated_completion = (
                        node.state_received_at
                        + datetime.timedelta(0, frame.remaining_time)
                    )
                    PYVLXLOG.debug(
                        "%s will be closing until", node.estimated_completion
                    )
                else:
                    if node.is_opening:
                        node.is_opening = False
                        node.state_received_at = None
                        node.estimated_completion = None
                        PYVLXLOG.debug("%s stops opening", node.name)
                    if node.is_closing:
                        node.is_closing = False
                        PYVLXLOG.debug("%s stops closing", node.name)

            # Set main parameter
            if isinstance(node, OpeningDevice):
                if position.position <= Parameter.MAX:
                    node.position = position
                    node.target = target
                    PYVLXLOG.debug("%s position changed to: %s", node.name, position)
                await node.after_update()
            elif isinstance(node, DimmableDevice):
                intensity = Intensity(frame.current_position)
                if intensity.intensity <= Parameter.MAX:
                    node.intensity = intensity
                    PYVLXLOG.debug("%s intensity changed to: %s", node.name, intensity)
                await node.after_update()
            elif isinstance(node, OnOffSwitch):
                state = SwitchParameter(frame.current_position)
                target = SwitchParameter(frame.target)
                if state.state == target.state:
                    if state.state == Parameter.ON:
                        node.parameter = state
                        PYVLXLOG.debug("%s state changed to: %s", node.name, state)
                    elif state.state == Parameter.OFF:
                        node.parameter = state
                        PYVLXLOG.debug("%s state changed to: %s", node.name, state)
                    await node.after_update()
        elif isinstance(frame, FrameStatusRequestNotification):
            await self.process_frame_status_request_notification(frame)



"""Unit tests for FrameDiscoverNodesConfirmation."""
import unittest

from pyvlx.api.frame_creation import frame_from_raw
from pyvlx.api.frames import FrameDiscoverNodesConfirmation


class TestFrameNodeDiscoverConfirmation(unittest.TestCase):
    """Test class for FrameDiscoverNodesConfirmation."""

    # pylint: disable=too-many-public-methods,invalid-name

    def test_bytes(self):
        """Test FrameDiscoverNodesConfirmation."""
        frame = FrameDiscoverNodesConfirmation()
        self.assertEqual(bytes(frame), b"\x00\x03\x01\x04\x06")

    def test_frame_from_raw(self):
        """Test parse FrameDiscoverNodesConfirmation from raw."""
        frame = frame_from_raw(b"\x00\x03\x01\x04\x06")
        self.assertTrue(isinstance(frame, FrameDiscoverNodesConfirmation))

    def test_str(self):
        """Test string representation of FrameDiscoverNodesConfirmation."""
        frame = FrameDiscoverNodesConfirmation()
        self.assertEqual(str(frame), "<FrameDiscoverNodesConfirmation/>")
